
from twisted.trial import unittest
from twisted.python.components import registerAdapter

from axiom.item import Item
from axiom.store import Store
from axiom.iaxiom import IPowerupIndirector
from axiom.attributes import integer, inmemory, reference

from zope.interface import Interface, implements, Attribute


class IValueHaver(Interface):
    value = Attribute("""
    An integer that you can add to other integers.
    """)

class IScalingFactor(Interface):
    scale = Attribute("""
    An integer that a sum can be multiplied by.
    """)

class ISumProducer(Interface):

    def doSum():
        """
        Produce a sum.
        """


class SumContributor(Item):
    schemaVersion = 1
    typeName = 'test_sum_contributor'

    value = integer()

class MinusThree(object):
    implements(IValueHaver)

    def __init__(self, otherValueHaver):
        self.value = otherValueHaver.value - 3

class SubtractThree(Item):
    schemaVersion = 1
    typeName = 'test_powerup_indirection_subtractthree'
    valueHaver = reference()

    implements(IPowerupIndirector)

    def indirect(self, iface):
        assert iface is IValueHaver, repr(iface)
        return MinusThree(self.valueHaver)


class PlusTwo(Item):
    """
    Example powerup with installation information.
    """
    implements(IValueHaver)
    powerupInterfaces = (IValueHaver,)

    value = integer(default=2)

class PlusOneTimesFour(Item):
    """
    Example powerup with dynamic installation information.
    """
    implements(IScalingFactor, IValueHaver)
    scale = integer(default=1)
    value = integer(default=4)

    def __getPowerupInterfaces__(self, powerup):
        yield (IScalingFactor, 1)
        yield (IValueHaver, 3)

class Summer(Item):
    schemaVersion = 1
    typeName = 'test_sum_doer'

    sumTimes = integer()
    sumTotal = integer()

    def __init__(self, **kw):
        super(Summer, self).__init__(**kw)
        self.sumTotal = 0
        self.sumTimes = 0

    def doSum(self):
        total = 0
        for haver in self.store.powerupsFor(IValueHaver):
            value = haver.value
            self.sumTotal += value
            total += value
        self.sumTimes += 1
        for factor in self.store.powerupsFor(IScalingFactor):
            value = factor.scale
            self.sumTotal *= value
            total *= value
        return total

class BrokenPowerup(Item):
    stuff = integer()

    def __getPowerupInterfaces__(self, pifs):
        return 'not a list of pairs'


class PowerUpTest(unittest.TestCase):

    def testBasicPowerups(self):
        # tests an interaction between __conform__ and other stuff

        s = Store()
        mm = Summer(store=s)
        s.powerUp(mm, ISumProducer)

        s.powerUp(SumContributor(store=s, value=1), IValueHaver)
        s.powerUp(SumContributor(store=s, value=2), IValueHaver)
        s.powerUp(SumContributor(store=s, value=3), IValueHaver)

        self.assertEquals(mm.doSum(), 6)

        s.close()


    def testPowerupIdentity(self):
        s = Store()
        mm = Summer(store=s)
        s.powerUp(mm, ISumProducer)

        sc3 = SumContributor(store=s, value=3)

        s.powerUp(SumContributor(store=s, value=1), IValueHaver)
        s.powerUp(SumContributor(store=s, value=2), IValueHaver)
        s.powerUp(sc3, IValueHaver)
        s.powerUp(sc3, IValueHaver)

        self.assertEquals(mm.doSum(), 6)

        s.close()



    def test_automaticPowerupInstall(self):
        """
        Powerups with 'powerupInterfaces' attributes can be installed
        on those interfaces without the caller needing to refer to
        them directly.
        """
        s = Store()
        mm = Summer(store=s)
        s.powerUp(mm, ISumProducer)
        p = PlusTwo(store=s)
        s.powerUp(p)

        self.assertEquals(mm.doSum(), 2)


    def test_dynamicAutomaticPowerupInstall(self):
        """
        Powerups with '__getPowerupInterfaces__' methods can be
        installed on the interfaces in the iterable that method
        returns.
        """
        s = Store()
        mm = Summer(store=s)
        s.powerUp(mm, ISumProducer)
        p = PlusOneTimesFour(store=s)
        s.powerUp(p)

        self.assertEquals(mm.doSum(), 4)


    def test_dynamicAutomaticPowerupFailure(self):
        """
        Powerups with '__getPowerupInterfaces__' methods that don't return
        iterables of pairs report an informative error message when powered up.
        """
        s = Store()
        mm = Summer(store=s)
        s.powerUp(mm, ISumProducer)
        p = BrokenPowerup(store=s)
        err = self.assertRaises(ValueError, s.powerUp, p)
        self.assertEquals(str(err),
                          'return value from %r.__getPowerupInterfaces__'
                          ' not an iterable of 2-tuples' % (p,))


    def test_automaticPowerDown(self):
        """
        Powerups with 'powerupInterfaces' attributes can be powered
        down automatically on the interfaces they specify.
        """
        s = Store()
        p = PlusTwo(store=s)
        s.powerUp(p)
        s.powerDown(p)
        self.assertEquals(len(list(s.powerupsFor(IValueHaver))), 0)

    def test_automaticDynamicPowerDown(self):
        """
        Powerups with '__getPowerupInterfaces__' methods can be powered
        down automatically on the interfaces they specify.
        """
        s = Store()
        p = PlusOneTimesFour(store=s)
        s.powerUp(p)
        s.powerDown(p)
        self.assertEquals(len(list(s.powerupsFor(IValueHaver))), 0)
        self.assertEquals(len(list(s.powerupsFor(IScalingFactor))), 0)

    def testIndirectedPowerups(self):
        """
        Powerups which implement L{IPowerupIndirector} should not be returned
        directly, the values that they return from indirect() should be
        returned directly.
        """
        s = Store()
        mm = Summer(store=s)
        s.powerUp(
            SubtractThree(
                store=s, valueHaver=SumContributor(store=s, value=5)),
            IValueHaver)
        self.assertEquals(mm.doSum(), 2)
        s.close()


    def testNoIndirectedIndirection(self):
        """
        Because it is a special interface in the powerup system, you can't have
        powerups for IPowerupIndirector; there's no sensible thing that could
        mean other than an infinite loop. Let's make sure that both looking for
        IPowerupIndirector and attempting to install a powerup for it will fail
        appropriately.
        """
        s = Store()
        s3 = SubtractThree(store=s)
        self.assertRaises(TypeError, s.powerUp, s3, IPowerupIndirector)
        self.assertEqual(list(s.powerupsFor(IPowerupIndirector)), [])



from twisted.application.service import IService, Service

class SillyService(Item, Service):
    typeName = 'test_silly_service'
    schemaVersion = 1

    started = integer(default=0)
    stopped = integer(default=0)
    running = integer(default=0)

    parent = inmemory()

    def startService(self):
        self.started += 1
        self.running = 1

    def stopService(self):
        assert self.running
        self.running = 0
        self.stopped += 1

class SpecialCaseTest(unittest.TestCase):

    def testStoreServicePowerup(self):
        s = Store()
        ss = SillyService(store=s)
        s.powerUp(ss, IService)
        IService(s).startService()
        IService(s).stopService()
        self.assertEquals(ss.started, 1)
        self.assertEquals(ss.stopped, 1)
        self.assertEquals(ss.running, 0)

    def testItemServicePowerup(self):
        s = Store()
        sm = Summer(store=s)
        ss = SillyService(store=s)
        sm.powerUp(ss, IService)
        IService(sm).startService()
        IService(sm).stopService()
        self.assertEquals(ss.started, 1)
        self.assertEquals(ss.stopped, 1)
        self.assertEquals(ss.running, 0)



class ItemWithAdapter(Item):
    """
    An item which will have an adapter registered for its type.
    """
    attribute = integer()

registerAdapter(lambda o: 42, ItemWithAdapter, ISumProducer)



class InMemoryPowerupTests(unittest.TestCase):
    """
    Tests for the behavior of powerups which are not database-resident.
    """
    def _createEmpowered(self, withStore=True):
        powerup = object()
        item = SumContributor(store=Store() if withStore else None)
        item.inMemoryPowerUp(powerup, ISumProducer)
        return powerup, item


    def test_powerupsFor(self):
        """
        L{Item.powerupsFor} returns a list the first element of which is the
        object previously passed to L{Item.inMemoryPowerUp}.
        """
        powerup, item = self._createEmpowered()
        self.assertEqual(list(item.powerupsFor(ISumProducer)), [powerup])


    def test_inMemoryPriority(self):
        """
        Adapting an L{Item} to an interface results in the in-memory powerup on
        that item for that interface even if there are database-resident
        powerups on that item for that interface.
        """
        powerup, item = self._createEmpowered()
        item.powerUp(item, ISumProducer)
        self.assertIdentical(ISumProducer(item), powerup)


    def test_conformWithoutStore(self):
        """
        Adaptation (through L{Item.__conform__}) should be allowed even if the
        object is not stored, as long as it has an in-memory powerup.
        """
        powerup, item = self._createEmpowered(withStore=False)
        self.assertIdentical(ISumProducer(item), powerup)


    def test_noPowerups(self):
        """
        L{Item.powerupsFor} returns no powerups for a storeless item with no
        powerups for the given interface, and adaption to the interface fails
        with L{TypeError}.
        """
        item = SumContributor()
        self.assertEquals(list(item.powerupsFor(ISumProducer)), [])
        self.assertRaises(TypeError, ISumProducer, item)


    def test_adapterNoPowerups(self):
        """
        Adapting an item to an interface for which no powerups are installed
        will allow adaption to proceed via a registered adapter.
        """
        item = ItemWithAdapter()
        self.assertEquals(ISumProducer(item), 42)




from axiom.attributes import integer, reference
from axiom.item import Item, normalize
from axiom.upgrade import registerUpgrader

class Referrer(Item):
    # Don't import the old schema. -exarkun
    typeName = normalize(
        'axiom.test.upgrade_fixtures.two_upgrades_old.Referrer')
    schemaVersion = 2
    referee = reference()

def upgradeReferrer1to2(old):
    return old.upgradeVersion(
        old.typeName, 1, 2,
        referee=old.referee)

registerUpgrader(upgradeReferrer1to2, Referrer.typeName, 1, 2)

class Referee(Item):
    # Don't import the old schema. -exarkun
    typeName = normalize(
        'axiom.test.upgrade_fixtures.two_upgrades_old.Referee')
    schemaVersion = 2
    dummy = integer()

def upgradeReferee1to2(old):
    return old.upgradeVersion(
        old.typeName, 1, 2,
        dummy=old.dummy)

registerUpgrader(upgradeReferee1to2, Referee.typeName, 1, 2)




"""
Tests for the upgrade of SubScheduler from version 1 to version 2, in which it
was largely supplanted by L{_UserScheduler}.
"""

from axiom.iaxiom import IScheduler
from axiom.substore import SubStore
from axiom.scheduler import SubScheduler, _UserScheduler
from axiom.test.historic.stubloader import StubbedTest


class SubSchedulerUpgradeTests(StubbedTest):
    def test_powerdown(self):
        """
        The L{SubScheduler} created by the stub is powered down by the upgrade
        and adapting the L{Store} to L{IScheduler} succeeds with a
        L{_UserScheduler}.
        """
        sub = self.store.findFirst(SubStore).open()
        upgraded = sub.whenFullyUpgraded()
        def subUpgraded(ignored):
            scheduler = sub.findUnique(SubScheduler)
            self.assertEquals(list(sub.interfacesFor(scheduler)), [])

            self.assertIsInstance(IScheduler(sub), _UserScheduler)
        upgraded.addCallback(subUpgraded)
        return upgraded



# -*- test-case-name: axiom.test.test_upgrading.SwordUpgradeTest.test_upgradeSkipVersion -*-

"""
This is a newer version of the module found in oldapp.py, used in the
upgrading tests.  It upgrades from oldapp in one step, rather than requiring an
intermediary step as morenewapp.py does.
"""

from axiom.item import Item
from axiom.attributes import text, integer, reference, inmemory

from axiom.upgrade import registerUpgrader

class ActivateHelper:
    activated = 0
    def activate(self):
        self.activated += 1

class Adventurer(ActivateHelper, Item):
    typeName = 'test_app_player'
    schemaVersion = 2

    name = text()
    activated = inmemory()

class InventoryEntry(ActivateHelper, Item):
    typeName = 'test_app_inv'
    schemaVersion = 1

    owner = reference()
    owned = reference()

    activated = inmemory()

class Sword(ActivateHelper, Item):
    typeName = 'test_app_sword'
    schemaVersion = 3

    name = text()
    damagePerHit = integer()
    activated = inmemory()

    def owner():
        def get(self):
            return self.store.findUnique(InventoryEntry,
                                         InventoryEntry.owned == self).owner
        return get,

    owner = property(*owner())


def sword2to3(oldsword):
    raise RuntimeError("The database does not contain any swords of version 2,"
                       " so you should be able to skip this version.")


registerUpgrader(sword2to3, 'test_app_sword', 2, 3)


####### DOUBLE-LEGACY UPGRADE SPECTACULAR !! ###########

# declare legacy class.

from axiom.item import declareLegacyItem

declareLegacyItem(typeName = 'test_app_sword',
                  schemaVersion = 2,
                  attributes = dict(name=text(),
                                    damagePerHit=integer(),
                                    owner=reference(),
                                    activated=inmemory()))


def upgradePlayerAndSword(oldplayer):
    newplayer = oldplayer.upgradeVersion('test_app_player', 1, 2)
    newplayer.name = oldplayer.name

    oldsword = oldplayer.sword

    newsword = oldsword.upgradeVersion('test_app_sword', 1, 3,
                                       name=oldsword.name,
                                       damagePerHit=oldsword.hurtfulness * 2)
    invent = InventoryEntry(store=newsword.store,
                            owner=newplayer,
                            owned=newsword)
    return newplayer, newsword


def player1to2(oldplayer):
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newplayer


def sword1to3(oldsword):
    oldPlayerType = oldsword.store.getOldVersionOf('test_app_player', 1)
    oldplayer = list(oldsword.store.query(oldPlayerType,
                                          oldPlayerType.sword == oldsword))[0]
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newsword


registerUpgrader(sword1to3, 'test_app_sword', 1, 3)
registerUpgrader(player1to2, 'test_app_player', 1, 2)




# -*- test-case-name: axiom.test.test_listversions -*-

from zope.interface import classProvides
from twisted import plugin
from twisted.python import usage, versions
from axiom import iaxiom, item, attributes, plugins
from axiom.scripts import axiomatic
from epsilon.extime import Time


class ListVersions(usage.Options, axiomatic.AxiomaticSubCommandMixin):
    """
    Command for listing the version history of a store.
    """

    classProvides(plugin.IPlugin, iaxiom.IAxiomaticCommand)
    name = "list-version"
    description = "Display software package version history."

    def postOptions(self):
        for line in listVersionHistory(self.parent.getStore()):
            print line



class SystemVersion(item.Item):
    """
    Represents a set of software package versions which, taken together,
    comprise a "system version" of the software that can have affected
    the contents of a Store.

    By recording the changes of these versions in the store itself we can
    better reconstruct its history later.
    """

    creation = attributes.timestamp(
        doc="When this system version set was recorded.",
        allowNone=False)


    def __repr__(self):
        return '<SystemVersion %s>' % (self.creation,)


    def longWindedRepr(self):
        """
        @return: A string representation of this SystemVersion suitable for
        display to the user.
        """
        return '\n\t'.join(
            [repr(self)] + [repr(sv) for sv in self.store.query(
                        SoftwareVersion,
                        SoftwareVersion.systemVersion == self)])



class SoftwareVersion(item.Item):
    """
    An Item subclass to map L{twisted.python.versions.Version} objects.
    """

    systemVersion = attributes.reference(
        doc="The system version this package version was observed in.",
        allowNone=False)

    package = attributes.text(doc="The software package.",
                              allowNone=False)
    version = attributes.text(doc="The version string of the software.",
                              allowNone=False)
    major = attributes.integer(doc='Major version number.',
                               allowNone=False)
    minor = attributes.integer(doc='Minor version number.',
                               allowNone=False)
    micro = attributes.integer(doc='Micro version number.',
                               allowNone=False)


    def asVersion(self):
        """
        Convert the version data in this item to a
        L{twisted.python.versions.Version}.
        """
        return versions.Version(self.package, self.major, self.minor, self.micro)


    def __repr__(self):
        return '<SoftwareVersion %s: %s>' % (self.package, self.version)



def makeSoftwareVersion(store, version, systemVersion):
    """
    Return the SoftwareVersion object from store corresponding to the
    version object, creating it if it doesn't already exist.
    """
    return store.findOrCreate(SoftwareVersion,
                              systemVersion=systemVersion,
                              package=unicode(version.package),
                              version=unicode(version.short()),
                              major=version.major,
                              minor=version.minor,
                              micro=version.micro)



def listVersionHistory(store):
    """
    List the software package version history of store.
    """
    q = store.query(SystemVersion, sort=SystemVersion.creation.descending)
    return [sv.longWindedRepr() for sv in q]


def getSystemVersions(getPlugins=plugin.getPlugins):
    """
    Collect all the version plugins and extract their L{Version} objects.
    """
    return list(getPlugins(iaxiom.IVersion, plugins))


def checkSystemVersion(s, versions=None):
    """
    Check if the current version is different from the previously recorded
    version.  If it is, or if there is no previously recorded version,
    create a version matching the current config.
    """

    if versions is None:
        versions = getSystemVersions()

    currentVersionMap = dict([(v.package, v) for v in versions])
    mostRecentSystemVersion = s.findFirst(SystemVersion,
                                          sort=SystemVersion.creation.descending)
    mostRecentVersionMap = dict([(v.package, v.asVersion()) for v in
                                 s.query(SoftwareVersion,
                                         (SoftwareVersion.systemVersion ==
                                          mostRecentSystemVersion))])

    if mostRecentVersionMap != currentVersionMap:
        currentSystemVersion = SystemVersion(store=s, creation=Time())
        for v in currentVersionMap.itervalues():
            makeSoftwareVersion(s, v, currentSystemVersion)




"""
Benchmark the tagNames method of L{axiom.tags.Catalog}
"""

import time, sys

from epsilon.scripts import benchmark

from axiom import store, item, attributes, tags

N_TAGS = 20
N_COPIES = 5000
N_LOOPS = 1000

class TaggedObject(item.Item):
    name = attributes.text()



def main():
    s = store.Store("tags.axiom")
    c = tags.Catalog(store=s)
    o = TaggedObject(store=s)

    def tagObjects(tag, copies):
        for x in xrange(copies):
            c.tag(o, tag)
    for i in xrange(N_TAGS):
        s.transact(tagObjects, unicode(i), N_COPIES)

    def getTags():
        for i in xrange(N_LOOPS):
            list(c.tagNames())

    benchmark.start()
    s.transact(getTags)
    benchmark.stop()



if __name__ == '__main__':
    main()



from twisted.trial.unittest import TestCase

class UnavailableTypeTestCase(TestCase):
    def testUnavailable(self):
        from axiom import attributes, item, store

        def makeItem():
            class MyItem(item.Item):

                typeName = 'test_deadtype_myitem'
                schemaVersion = 1

                hello = attributes.integer()

            return MyItem

        storedir = self.mktemp()

        theStore = store.Store(storedir)
        makeItem()(store=theStore)

        item = reload(item)
        store = reload(store)

        store.Store(storedir)

    testUnavailable.skip = 'This test breaks EVERY subsequent test, because reloading item and store is not allowed'



# -*- test-case-name: axiom.test.test_upgrading.SchemaUpgradeTest.testUpgradeWithMissingVersion -*-


from axiom.item import Item
from axiom.attributes import text, integer, reference, inmemory

from axiom.upgrade import registerUpgrader

class ActivateHelper:
    activated = 0
    def activate(self):
        self.activated += 1

class Adventurer(ActivateHelper, Item):
    typeName = 'test_app_player'
    schemaVersion = 2

    name = text()
    activated = inmemory()

class InventoryEntry(ActivateHelper, Item):
    typeName = 'test_app_inv'
    schemaVersion = 1

    owner = reference()
    owned = reference()

    activated = inmemory()

class Sword(ActivateHelper, Item):
    typeName = 'test_app_sword'
    schemaVersion = 3

    name = text()
    damagePerHit = integer()
    activated = inmemory()

    def owner():
        def get(self):
            return self.store.findUnique(InventoryEntry,
                                         InventoryEntry.owned == self).owner
        return get,

    owner = property(*owner())


def sword2to3(oldsword):
    newsword = oldsword.upgradeVersion('test_app_sword', 2, 3)
    n = oldsword.store.getOldVersionOf('test_app_sword', 2)
    itrbl = oldsword.store.query(n)
    newsword.name = oldsword.name
    newsword.damagePerHit = oldsword.damagePerHit
    invent = InventoryEntry(store=newsword.store,
                            owner=oldsword.owner,
                            owned=newsword)
    return newsword


registerUpgrader(sword2to3, 'test_app_sword', 2, 3)

# ...

# deliberately missing upgraders for faulty upgrade path tests: see
# morenewapp.py for multi-step upgrades.



# -*- test-case-name: axiom.test.test_upgrading.DuringUpgradeTests.test_reentrantUpgraderFailure -*-

from axiom.attributes import integer
from axiom.item import Item

class Simple(Item):
    """
    A simple item that doesn't do much.
    """
    dummy = integer()




from axiom.substore import SubStore
from axiom.test.historic.stubloader import StubbedTest

from axiom.test.historic.stub_subStoreStartupService1to2 import DummyService

class UpgradeTest(StubbedTest):
    def testSubStoreServiceStarterStoppedStartingSubStoreServices(self):
        """
        Verify that the sub-store service starter is removed and substore services
        will not be started.

        Also, say that nine times fast.
        """
        ss = self.store.findUnique(SubStore)
        thePowerup = ss.open().findUnique(DummyService)

        # The upgrade stub-loading framework actually takes care of invoking
        # the parent store startService, so we don't have to do that here;
        # let's just make sure that the substore's service wasn't started as
        # part of the upgrade.
        self.failIf(thePowerup.everStarted)



# -*- test-case-name: axiom.test.test_upgrading.DuringUpgradeTests.test_referenceModifiedByForeignUpgrader -*-

from axiom.attributes import reference, integer
from axiom.item import Item

OLD_VALUE = 69

class Referrer(Item):
    """
    An item which just refers to another kind of item which will be upgraded.
    """
    referee = reference()


class Referee(Item):
    """
    An item the upgrader of which replaces itself on L{Referrer} with a new
    instance with a different value.
    """
    value = integer()



# -*- test-case-name: axiom.test.test_upgrading.DeletionTest.testCircular -*-
from axiom.item import Item

from axiom.attributes import reference, integer

class A(Item):
    typeName = 'test_circular_a'
    b = reference()

class B(Item):
    typeName = 'test_circular_b'
    a = reference()
    n = integer()

    schemaVersion = 2

from axiom.upgrade import registerUpgrader

def b1to2(oldb):
    # This upgrader isn't doing anything that actually makes sense; in a
    # realistic upgrader, you'd probably be changing A around, perhaps deleting
    # it to destroy old adjunct items and creating a new A.  The point is,
    # s.findUnique(A).b should give back the 'b' that you are upgrading whether
    # it is run before or after the upgrade.
    oldb.a.deleteFromStore()
    newb = oldb.upgradeVersion('test_circular_b', 1, 2)
    newb.n = oldb.n
    newb.a = A(store=newb.store,
               b=newb)
    return newb

registerUpgrader(b1to2, 'test_circular_b', 1, 2)




from twisted.trial.unittest import TestCase

from axiom.item import Item
from axiom.attributes import integer

from axiom.slotmachine import hyper as super

__metaclass__ = type

class X:
    xm = 0
    def m(self):
        self.xm += 1
        return self.xm

class Y(X):
    ym = 0

    def m(self):
        ret = super(Y, self).m()
        self.ym += 1
        ret += 1
        return ret

class Z(X):
    zm = 0
    def m(self):
        ret = super(Z, self).m()
        ret += 1
        self.zm += 1
        return ret

class XYZ(Y, Z):
    pass

class ItemXYZ(Item, XYZ):
    typeName = 'item_xyz'
    schemaVersion = 1

    xm = integer(default=0)
    ym = integer(default=0)
    zm = integer(default=0)


class TestBorrowedMixins(TestCase):

    def testSanity(self):
        xyz = XYZ()
        val = xyz.m()
        self.assertEquals(val, 3)

    def testItemSanity(self):
        xyz = ItemXYZ()
        val = xyz.m()
        self.assertEquals(val, 3)



# -*- test-case-name: axiom.test -*-
from axiom._version import __version__
from twisted.python import versions

def asTwistedVersion(packageName, versionString):
    return versions.Version(packageName, *map(int, versionString.split(".")[:3]))

version = asTwistedVersion("axiom", __version__)




from axiom.store import Store
from axiom.substore import SubStore
from axiom.item import Item

from axiom.attributes import integer

from twisted.trial.unittest import TestCase
from twisted.python import filepath

class ExplosiveItem(Item):

    nothing = integer()

    def yourHeadAsplode(self):
        1 / 0


class CrossStoreTest(TestCase):

    def setUp(self):
        self.spath = filepath.FilePath(self.mktemp() + ".axiom")
        self.store = Store(self.spath)
        self.substoreitem = SubStore.createNew(self.store,
                                               ["sub.axiom"])

        self.substore = self.substoreitem.open()
        # Not available yet.
        self.substore.attachToParent()


class TestCrossStoreTransactions(CrossStoreTest):

    def testCrossStoreTransactionality(self):
        def createTwoSubStoreThings():
            ExplosiveItem(store=self.store)
            ei = ExplosiveItem(store=self.substore)
            ei.yourHeadAsplode()

        self.failUnlessRaises(ZeroDivisionError,
                              self.store.transact,
                              createTwoSubStoreThings)

        self.failUnlessEqual(
            self.store.query(ExplosiveItem).count(),
            0)

        self.failUnlessEqual(
            self.substore.query(ExplosiveItem).count(),
            0)

class TestCrossStoreInsert(CrossStoreTest):

    def testCrossStoreInsert(self):
        def populate(s, n):
            for i in xrange(n):
                ExplosiveItem(store=s)

        self.store.transact(populate, self.store, 2)
        self.store.transact(populate, self.substore, 3)

        self.failUnlessEqual(
            self.store.query(ExplosiveItem).count(),
            2)

        self.failUnlessEqual(
            self.substore.query(ExplosiveItem).count(),
            3)



# -*- test-case-name: axiom.test.historic.test_catalog1to2 -*-


from axiom.item import Item
from axiom.attributes import text
from axiom.tags import Catalog
from axiom.test.historic.stubloader import saveStub

class Dummy(Item):
    attribute = text(doc="dummy attribute")



def createDatabase(s):
    """
    Populate the given Store with a Catalog and some Tags.
    """
    c = Catalog(store=s)
    c.tag(c, u"internal")
    c.tag(s, u"internal")
    i = Dummy(store=s)
    c.tag(i, u"external")
    c.tag(i, u"green")



if __name__ == '__main__':
    saveStub(createDatabase, 6917)



# Copyright (c) 2008 Divmod.  See LICENSE for details.

"""
Plugins provided by Axiom for Axiom.
"""

import getpass
import code, os, traceback, sys
try:
    import readline
except ImportError:
    readline = None

from zope.interface import directlyProvides

from twisted.python import usage, filepath, log
from twisted.python.reflect import qual
from twisted.plugin import IPlugin

from epsilon.hotfix import require
require('twisted', 'filepath_copyTo')

import axiom
from axiom import store, attributes, userbase, dependency, errors
from axiom.substore import SubStore
from axiom.scripts import axiomatic
from axiom.listversions import ListVersions
from axiom import version
from axiom.iaxiom import IVersion
from axiom.upgrade import upgradeExplicitOid

directlyProvides(version, IPlugin, IVersion)

#placate pyflakes
ListVersions



class Upgrade(axiomatic.AxiomaticCommand):
    name = 'upgrade'
    description = 'Synchronously upgrade an Axiom store and substores'

    optParameters = [
        ('count', 'n', '100', 'Number of upgrades to perform per transaction')]

    errorMessageFormat = 'Error upgrading item (with typeName=%s and storeID=%d) from version %d to %d.'

    def upgradeEverything(self, store):
        """
        Upgrade all the items in C{store}.
        """
        for dummy in store._upgradeManager.upgradeBatch(self.count):
            pass


    def upgradeStore(self, store):
        """
        Recursively upgrade C{store}.
        """
        self.upgradeEverything(store)
        upgradeExplicitOid(store)

        for substore in store.query(SubStore):
            print 'Upgrading: {!r}'.format(substore)
            self.upgradeStore(substore.open())

    def perform(self, store, count):
        """
        Upgrade C{store} performing C{count} upgrades per transaction.

        Also, catch any exceptions and print out something useful.
        """
        self.count = count

        try:
            self.upgradeStore(store)
            print 'Upgrade complete'
        except errors.ItemUpgradeError as e:
            print 'Upgrader error:'
            e.originalFailure.printTraceback(file=sys.stdout)
            print self.errorMessageFormat % (
                e.oldType.typeName, e.storeID, e.oldType.schemaVersion,
                e.newType.schemaVersion)


    def postOptions(self):
        try:
            count = int(self['count'])
        except ValueError:
            raise usage.UsageError('count must be an integer')

        siteStore = self.parent.getStore()
        self.perform(siteStore, count)



class AxiomConsole(code.InteractiveConsole):
    def runcode(self, code):
        """
        Override L{code.InteractiveConsole.runcode} to run the code in a
        transaction unless the local C{autocommit} is currently set to a true
        value.
        """
        if not self.locals.get('autocommit', None):
            return self.locals['db'].transact(code.InteractiveConsole.runcode, self, code)
        return code.InteractiveConsole.runcode(self, code)



class Browse(axiomatic.AxiomaticCommand):
    synopsis = "[options]"

    name = 'browse'
    description = 'Interact with an Axiom store.'

    optParameters = [
        ('history-file', 'h', '~/.axiomatic-browser-history',
         'Name of the file to which to save input history.'),
        ]

    optFlags = [
        ('debug', 'b', 'Open Store in debug mode.'),
        ]

    def postOptions(self):
        interp = code.InteractiveConsole(self.namespace(), '<axiom browser>')
        historyFile = os.path.expanduser(self['history-file'])
        if readline is not None and os.path.exists(historyFile):
            readline.read_history_file(historyFile)
        try:
            interp.interact("%s.  Autocommit is off." % (str(axiom.version),))
        finally:
            if readline is not None:
                readline.write_history_file(historyFile)


    def namespace(self):
        """
        Return a dictionary representing the namespace which should be
        available to the user.
        """
        self._ns = {
            'db': self.store,
            'store': store,
            'autocommit': False,
            }
        return self._ns



class UserbaseMixin:
    def installOn(self, other):
        # XXX check installation on other, not store
        for ls in self.store.query(userbase.LoginSystem):
            raise usage.UsageError("UserBase already installed")
        else:
            ls = userbase.LoginSystem(store=self.store)
            dependency.installOn(ls, other)
            return ls



class Install(axiomatic.AxiomaticSubCommand, UserbaseMixin):
    def postOptions(self):
        self.installOn(self.store)



class Create(axiomatic.AxiomaticSubCommand, UserbaseMixin):
    synopsis = "<username> <domain> [password]"

    def parseArgs(self, username, domain, password=None):
        self['username'] = self.decodeCommandLine(username)
        self['domain'] = self.decodeCommandLine(domain)
        self['password'] = password


    def postOptions(self):
        msg = 'Enter new AXIOM password: '
        while not self['password']:
            password = getpass.getpass(msg)
            second = getpass.getpass('Repeat to verify: ')
            if password == second:
                self['password'] = password
            else:
                msg = 'Passwords do not match.  Enter new AXIOM password: '
        self.addAccount(
            self.store, self['username'], self['domain'], self['password'])


    def addAccount(self, siteStore, username, domain, password):
        """
        Create a new account in the given store.

        @param siteStore: A site Store to which login credentials will be
        added.
        @param username: Local part of the username for the credentials to add.
        @param domain: Domain part of the username for the credentials to add.
        @param password: Password for the credentials to add.
        @rtype: L{LoginAccount}
        @return: The added account.
        """
        for ls in siteStore.query(userbase.LoginSystem):
            break
        else:
            ls = self.installOn(siteStore)
        try:
            acc = ls.addAccount(username, domain, password)
        except userbase.DuplicateUser:
            raise usage.UsageError("An account by that name already exists.")
        return acc



class Disable(axiomatic.AxiomaticSubCommand):
    synopsis = "<username> <domain>"

    def parseArgs(self, username, domain):
        self['username'] = self.decodeCommandLine(username)
        self['domain'] = self.decodeCommandLine(domain)

    def postOptions(self):
        for acc in self.store.query(userbase.LoginAccount,
                                    attributes.AND(userbase.LoginAccount.username == self['username'],
                                                   userbase.LoginAccount.domain == self['domain'])):
            if acc.disabled:
                raise usage.UsageError("That account is already disabled.")
            else:
                acc.disabled = True
                break
        else:
            raise usage.UsageError("No account by that name exists.")



class List(axiomatic.AxiomaticSubCommand):
    def postOptions(self):
        acc = None
        for acc in self.store.query(userbase.LoginMethod):
            if acc.domain is None:
                print acc.localpart,
            else:
                print acc.localpart + '@' + acc.domain,
            if acc.account.disabled:
                print '[DISABLED]'
            else:
                print
        if acc is None:
            print 'No accounts'



class UserBaseCommand(axiomatic.AxiomaticCommand):
    name = 'userbase'
    description = 'LoginSystem introspection and manipulation.'

    subCommands = [
        ('install', None, Install, "Install UserBase on an Axiom database"),
        ('create', None, Create, "Create a new user"),
        ('disable', None, Disable, "Disable an existing user"),
        ('list', None, List, "List users in an Axiom database"),
        ]

    def getStore(self):
        return self.parent.getStore()



class Extract(axiomatic.AxiomaticCommand):
    name = 'extract-user'
    description = 'Remove an account from the login system, moving its associated database to the filesystem.'
    optParameters = [
        ('address', 'a', None, 'localpart@domain-format identifier of the user store to extract.'),
        ('destination', 'd', None, 'Directory into which to extract the user store.')]

    def extractSubStore(self, localpart, domain, destinationPath):
        siteStore = self.parent.getStore()
        la = siteStore.findFirst(
            userbase.LoginMethod,
            attributes.AND(userbase.LoginMethod.localpart == localpart,
                           userbase.LoginMethod.domain == domain)).account
        userbase.extractUserStore(la, destinationPath)


    def postOptions(self):
        localpart, domain = self.decodeCommandLine(self['address']).split('@', 1)
        destinationPath = filepath.FilePath(
            self.decodeCommandLine(self['destination'])).child(localpart + '@' + domain + '.axiom')
        self.extractSubStore(localpart, domain, destinationPath)



class Insert(axiomatic.AxiomaticCommand):
    name = 'insert-user'
    description = 'Insert a user store, such as one extracted with "extract-user", into a site store and login system.'
    optParameters = [
        ('userstore', 'u', None, 'Path to user store to be inserted.')
        ]

    def postOptions(self):
        userbase.insertUserStore(self.parent.getStore(),
                                 filepath.FilePath(self.decodeCommandLine(self['userstore'])))



"""
Hypothesis strategies for generating Axiom-related data.
"""
from epsilon.extime import Time
from hypothesis import strategies as st
from hypothesis.extra.datetime import datetimes

from axiom.attributes import LARGEST_NEGATIVE, LARGEST_POSITIVE



def axiomText(*a, **kw):
    """
    Strategy for generating Axiom-compatible text values.
    """
    return st.text(
        alphabet=st.characters(
            blacklist_categories={'Cs'},
            blacklist_characters={u'\x00'}),
        *a, **kw)


def textlists():
    """
    Strategy for generating lists storable with L{axiom.attributes.textlist}.
    """
    return st.lists(st.text(
        alphabet=st.characters(
            blacklist_categories={'Cs'},
            blacklist_characters={u'\x00', u'\x02', u'\x1f'})))



def axiomIntegers(minValue=LARGEST_NEGATIVE, maxValue=LARGEST_POSITIVE):
    """
    Strategy for generating Axiom-compatible integers.

    @type minValue: L{int}
    @param minValue: Minimum value to generate; default is the least value
        that can be stored in an L{axiom.attributes.integer} attribute.

    @type manValue: L{int}
    @param manValue: Maximum value to generate; default is the greatest value
        that can be stored in an L{axiom.attributes.integer} attribute.
    """
    return st.integers(min_value=minValue, max_value=maxValue)



def timestamps(*a, **kw):
    """
    Strategy for generating L{epsilon.extime.Time} objects.
    """
    return st.builds(Time.fromDatetime, datetimes(timezones=[], *a, **kw))



def fixedDecimals(precision, minValue=None, maxValue=None):
    """
    Strategy for generating L{decimal.Decimal} values of a fixed precision.

    @type precision: L{decimal.Decimal}
    @param precision: The precision to use; for example, C{Decimal('0.01')} for
        a L{axiom.attributes.point2decimal} attribute.

    @type minValue: L{decimal.Decimal}
    @param minValue: The minimum value to generate, or C{None} for the least
        possible.
    @type minValue: L{decimal.Decimal}
    @param minValue: The maximum value to generate, or C{None} for the greatest
        possible.
    """
    if minValue is None:
        minValue = LARGEST_NEGATIVE
    else:
        minValue = int(minValue / precision)
    if maxValue is None:
        maxValue = LARGEST_POSITIVE
    else:
        maxValue = int(maxValue / precision)
    return st.integers(min_value=minValue, max_value=maxValue).map(
        lambda v: v * precision)



__all__ = [
    'axiomText', 'axiomIntegers', 'fixedDecimals', 'timestamps', 'textlists']




from axiom.tags import Catalog
from axiom.test.historic.stubloader import StubbedTest

class CatalogUpgradeTest(StubbedTest):
    def testCatalogTagNames(self):
        """
        Test that the tagNames method of L{axiom.tags.Catalog} returns all the
        correct tag names.
        """
        c = self.store.findUnique(Catalog)
        self.assertEquals(
            sorted(c.tagNames()),
            [u"external", u"green", u"internal"])




# DELETE_OBJECT = 'DELETE FROM axiom_objects WHERE oid = ?'
CREATE_OBJECT = 'INSERT INTO *DATABASE*.axiom_objects (type_id) VALUES (?)'
CREATE_TYPE = 'INSERT INTO *DATABASE*.axiom_types (typename, module, version) VALUES (?, ?, ?)'

GET_TABLE_INFO = 'PRAGMA *DATABASE*.table_info(?)'


# The storeID for an object must be unique over the lifetime of the store.
# Since the storeID is allocated by inserting into axiom_objects, we use
# AUTOINCREMENT so that oids/rowids and thus storeIDs are never reused.

# The column is named "oid" instead of "storeID" for backwards compatibility
# with the implicit oid/rowid column in old Stores.
CREATE_OBJECTS = """
CREATE TABLE *DATABASE*.axiom_objects (
    oid INTEGER PRIMARY KEY AUTOINCREMENT,
    type_id INTEGER NOT NULL
        CONSTRAINT fk_type_id REFERENCES axiom_types(oid)
)
"""

CREATE_OBJECTS_IDX = """
CREATE INDEX *DATABASE*.axiom_objects_type_idx
    ON axiom_objects(type_id);
"""

CREATE_TYPES = """
CREATE TABLE *DATABASE*.axiom_types (
    oid INTEGER PRIMARY KEY AUTOINCREMENT,
    typename VARCHAR,
    module VARCHAR,
    version INTEGER
)
"""

CREATE_ATTRIBUTES = """
CREATE TABLE *DATABASE*.axiom_attributes (
    type_id INTEGER,
    row_offset INTEGER,
    indexed BOOLEAN,
    sqltype VARCHAR,
    allow_none BOOLEAN,
    pythontype VARCHAR,
    attribute VARCHAR,
    docstring TEXT
)
"""

BASE_SCHEMA = [
    CREATE_OBJECTS, CREATE_OBJECTS_IDX, CREATE_TYPES, CREATE_ATTRIBUTES]


TYPEOF_QUERY = """
SELECT *DATABASE*.axiom_types.typename, *DATABASE*.axiom_types.module, *DATABASE*.axiom_types.version
    FROM *DATABASE*.axiom_types, *DATABASE*.axiom_objects
    WHERE *DATABASE*.axiom_objects.oid = ?
        AND *DATABASE*.axiom_types.oid = *DATABASE*.axiom_objects.type_id
"""

HAS_SCHEMA_FEATURE = ("SELECT COUNT(oid) FROM *DATABASE*.sqlite_master "
                      "WHERE type = ? AND name = ?")

IDENTIFYING_SCHEMA = ('SELECT indexed, sqltype, allow_none, attribute '
                      'FROM *DATABASE*.axiom_attributes WHERE type_id = ? '
                      'ORDER BY row_offset')

ADD_SCHEMA_ATTRIBUTE = (
    'INSERT INTO *DATABASE*.axiom_attributes '
    '(type_id, row_offset, indexed, sqltype, allow_none, attribute, docstring, pythontype) '
    'VALUES (?, ?, ?, ?, ?, ?, ?, ?)')

ALL_TYPES = 'SELECT oid, module, typename, version FROM *DATABASE*.axiom_types'

LATEST_TYPES = 'SELECT typename, MAX(version) FROM *DATABASE*.axiom_types GROUP BY typename'

GET_GREATER_VERSIONS_OF_TYPE = ('SELECT version FROM *DATABASE*.axiom_types '
                                'WHERE typename = ? AND version > ?')

SCHEMA_FOR_TYPE = ('SELECT indexed, pythontype, attribute, docstring '
                   'FROM *DATABASE*.axiom_attributes '
                   'WHERE type_id = ?')

CHANGE_TYPE = 'UPDATE *DATABASE*.axiom_objects SET type_id = ? WHERE oid = ?'

APP_VACUUM = 'DELETE FROM *DATABASE*.axiom_objects WHERE (type_id == -1) AND (oid != (SELECT MAX(oid) from *DATABASE*.axiom_objects))'




"""
Test cases for PySQLite2-specific parts of the backend.
"""

from twisted.trial.unittest import TestCase

from axiom._pysqlite2 import OperationalError, Connection
from axiom.test.cursortest import ConnectionTestCaseMixin, StubConnection


class PySQLite2StubConnection(StubConnection):
    def timeout(self):
        raise OperationalError('database is locked')


class ConnectionTestCase(ConnectionTestCaseMixin, TestCase):
    expectedUnderlyingExceptionClass = OperationalError

    def createStubConnection(self, *a, **kw):
        return PySQLite2StubConnection(*a, **kw)


    def createAxiomConnection(self, underlyingConnection, *a, **kw):
        return Connection(underlyingConnection, *a, **kw)


    def createRealConnection(self):
        """
        Create a memory-backed connection for integration testing.
        """
        return Connection.fromDatabaseName(":memory:")



from axiom import item, attributes

class Bucket(item.Item):
    typeName = 'bucket'
    schemaVersion = 1

    name = attributes.text()

    def getstuff(self):
        for food in self.store.query(FoodItem,
                                     FoodItem.bucket == self,
                                     sort=FoodItem.deliciousness.descending):
            food.extra.what()


class FoodItem(item.Item):
    typeName = 'food'
    schemaVersion = 1

    bucket = attributes.reference()
    extra = attributes.reference()
    deliciousness = attributes.integer(indexed=True)

class Chicken(item.Item):
    typeName = 'chicken'
    schemaVersion = 1

    epistemologicalBasisForCrossingTheRoad = attributes.text()
    def what(self):
        print 'chicken!'

class Biscuit(item.Item):
    typeName = 'biscuit'
    schemaVersion = 1

    fluffiness = attributes.integer()
    def what(self):
        print 'biscuits!'


from axiom.store import Store

s = Store()

u = Bucket(name=u'whatever', store=s)
c = Chicken(epistemologicalBasisForCrossingTheRoad=u'extropian', store=s)
b = Biscuit(fluffiness=100, store=s)

FoodItem(store=s, deliciousness=3, extra=c, bucket=u)
FoodItem(store=s, deliciousness=4, extra=b, bucket=u)

u.getstuff()




"""
Test upgrading L{_SubSchedulerParentHook} from version 2 to 3.
"""

from axiom.test.historic.stubloader import StubbedTest

from axiom.scheduler import _SubSchedulerParentHook
from axiom.substore import SubStore
from axiom.dependency import _DependencyConnector


class SubSchedulerParentHookUpgradeTests(StubbedTest):
    """
    Test upgrading L{_SubSchedulerParentHook} from version 2 to 3.
    """
    def setUp(self):
        d = StubbedTest.setUp(self)
        def cbSetUp(ignored):
            self.hook = self.store.findUnique(_SubSchedulerParentHook)
        d.addCallback(cbSetUp)
        return d


    def test_attributesCopied(self):
        """
        The only attribute of L{_SubSchedulerParentHook} which still exists at
        the current version, version 4, C{subStore}, ought to have been
        copied over.
        """
        self.assertIdentical(
            self.hook.subStore, self.store.findUnique(SubStore))


    def test_uninstalled(self):
        """
        The record of the installation of L{_SubSchedulerParentHook} on the
        store is deleted in the upgrade to schema version 4.
        """
        self.assertEquals(
            list(self.store.query(
                    _DependencyConnector,
                    _DependencyConnector.installee == self.hook)),
            [])




"""
This module contains tests for C{__repr__} implementations within Axiom, to
make sure they contain enough information to be useful, and work when objects
are incompletely initialized.
"""

from axiom.iaxiom import IOrdering
from axiom.attributes import integer, text, reference
from axiom.item import Item
from axiom.store import Store

from twisted.trial.unittest import TestCase

class ReprTesterItemClass(Item):
    intattr = integer()
    txtattr = text()

class ReprTesterWithReference(Item):
    """
    Test fixture for testing 'reference' attributes.
    """
    refattr = reference()


class BasicInformation(TestCase):
    """
    Basic tests to verify that C{__repr__} implementations for various axiom
    objects provide enough information to debug them.
    """

    def test_storeID(self):
        """
        Verify that the storeID column tells you that it is a storeID, and who
        it belongs to.
        """
        R = repr(ReprTesterItemClass.storeID)
        self.assertIn('storeID', R)
        self.assertIn(ReprTesterItemClass.__name__, R)
        self.assertNotIn('intattr', R)


    def test_query(self):
        """
        Verify that queries tell you something about their target and
        comparison.
        """
        s = Store()
        R = repr(s.query(ReprTesterItemClass,
                         ReprTesterItemClass.intattr == 1))
        self.assertIn('intattr', R)
        self.assertIn(ReprTesterItemClass.__name__, R)


    def test_simpleOrdering(self):
        """
        Verify that ordering objects tell you something about their ordering.
        """
        R = repr(ReprTesterItemClass.intattr.ascending)
        self.assertIn("intattr", R)
        self.assertIn("asc", R.lower()) # leaving this a little open-ended so
                                        # that we can fiddle with case, ASC and
                                        # DESC vs. ascending and descending

    def test_complexOrdering(self):
        """
        Verify that complex orderings tell us about their component parts.
        """
        R = repr(IOrdering((ReprTesterItemClass.intattr.ascending,
                            ReprTesterItemClass.txtattr.descending)))
        self.assertIn(repr(ReprTesterItemClass.intattr.ascending), R)
        self.assertIn(repr(ReprTesterItemClass.txtattr.descending), R)


    def test_referenceAttribute(self):
        """
        Verify that repr()ing an object with reference attributes will show the ID.
        """
        s = Store()
        i1 = ReprTesterWithReference(store=s)
        i2 = ReprTesterWithReference(store=s)
        i1.refattr = i2
        R = repr(i1)
        self.assertIn("reference(%d)" % (i2.storeID,), R)


    def test_recursiveReferenceAttribute(self):
        """
        Verify that repr()ing an object with reference attributes that refer to
        themselves will not recurse.
        """
        s = Store()
        i1 = ReprTesterWithReference(store=s)
        i1.refattr = i1
        R = repr(i1)
        self.assertIn("reference(%d)" % (i1.storeID,), R)


    def test_unstoredReferenceAttribute(self):
        """
        Verify that repr()ing an object with reference attributes that refer to
        items not in a store does something reasonable.
        """
        i1 = ReprTesterWithReference()
        i2 = ReprTesterWithReference()
        i1.refattr = i2
        R = repr(i1)
        self.assertIn("reference(unstored@%d)" % (id(i2),), R)





"""
Tests for table-creation.
"""
from axiom import item
from axiom import attributes
from axiom import store

from twisted.trial.unittest import TestCase
from twisted.python import filepath

class A(item.Item):
    typeName = 'test_table_creator'
    schemaVersion = 1

    attr = attributes.integer(default=3)


class SomeError(Exception):
    """
    Dummy error for testing.
    """



def createAndRaise(s):
    """
    Create an A item, then raise a L{SomeError} (which will revert the
    transaction).  This is because there is no direct API for creating tables.
    """
    A(store=s)
    raise SomeError()


class TableCreationTest(TestCase):
    """
    Tests for table creation.
    """
    def test_committedTableCreation(self):
        """
        When tables are created in a transaction which is committed, they should
        persist in both Axiom's in-memory schema representation and within the
        on-disk SQL store.
        """
        storedir = filepath.FilePath(self.mktemp())
        s1 = store.Store(storedir)
        s1.transact(A, store=s1)
        self.assertIn(A, s1.typeToTableNameCache)
        s1.close()
        s2 = store.Store(storedir)
        self.assertIn(A, s2.typeToTableNameCache)
        s2.close()


    def test_revertedTableCreation(self):
        """
        When tables are created in a transaction which is reverted, they should
        persist in neither the SQL store nor the in-memory schema
        representation.
        """
        storedir = self.mktemp()
        s1 = store.Store(storedir)
        self.assertRaises(SomeError, s1.transact, createAndRaise, s1)
        self.assertNotIn(A, s1.typeToTableNameCache)
        s1.close()
        s2 = store.Store(storedir)
        self.assertNotIn(A, s2.typeToTableNameCache)


    def test_differentStoreTableCreation(self):
        """
        If two different stores are opened before a given table is created, and
        one creates it, this should be transparent to both item creation and
        queries made from either store.
        """
        storedir = self.mktemp()
        s1 = store.Store(storedir)
        s2 = store.Store(storedir)
        a1 = A(store=s1)
        a2 = A(store=s2)
        self.assertEquals(list(s1.query(
                    A, sort=A.storeID.ascending).getColumn("storeID")),
                          [a1.storeID, a2.storeID])
        self.assertEquals(list(s2.query(
                    A, sort=A.storeID.ascending).getColumn("storeID")),
                          [a1.storeID, a2.storeID])


    def test_dontReadTheSchemaSoMuch(self):
        """
        This is a regression test for a bug in Axiom where the schema was
        refreshed from SQL every time a table needed to be created, regardless
        of whether the schema needed to be refreshed or not.  In addition to
        being logically incorrect, this error severely hurt performance.

        The schema should only be re-read when a change is detected, by way of
        a table being created in two different Store objects, as in the test
        above in L{TableCreationTest.test_differentStoreTableCreation}.
        """
        s1 = store.Store(filepath.FilePath(self.mktemp()))
        def die():
            self.fail("schema refreshed unnecessarily called too much")
        s1._startup = die
        A(store=s1)




from twisted.internet.defer import Deferred

from axiom.test.historic.stubloader import StubbedTest
from axiom.test.historic.stub_processor1to2 import DummyProcessor

class ProcessorUpgradeTest(StubbedTest):
    def setUp(self):
        # Ick, we need to catch the run event of DummyProcessor, and I can't
        # think of another way to do it.
        self.dummyRun = DummyProcessor.run.im_func
        self.calledBack = Deferred()
        def dummyRun(calledOn):
            self.calledBack.callback(calledOn)
        DummyProcessor.run = dummyRun

        return StubbedTest.setUp(self)


    def tearDown(self):
        # Okay this is a pretty irrelevant method on a pretty irrelevant class,
        # but we'll fix it anyway.
        DummyProcessor.run = self.dummyRun

        return StubbedTest.tearDown(self)


    def test_pollingRemoval(self):
        """
        Test that processors lose their idleInterval but none of the rest of
        their stuff, and that they get scheduled by the upgrader so they can
        figure out what state they should be in.
        """
        proc = self.store.findUnique(DummyProcessor)
        self.assertEquals(proc.busyInterval, 100)
        self.failIfEqual(proc.scheduled, None)
        def assertion(result):
            self.assertEquals(result, proc)
        return self.calledBack.addCallback(assertion)




import sys, os

from zope.interface import Interface, implementer

from twisted.trial import unittest
from twisted.trial.unittest import TestCase, SynchronousTestCase
from twisted.internet import error, protocol, defer, reactor
from twisted.protocols import policies
from twisted.python import log, filepath

from axiom import store, item
from axiom.store import Store
from axiom.item import Item, declareLegacyItem, empowerment
from axiom.errors import ChangeRejected
from axiom.test import itemtest, itemtestmain
from axiom.attributes import integer, text, inmemory

class ProcessFailed(Exception):
    pass

class ProcessOutputCollector(protocol.ProcessProtocol, policies.TimeoutMixin):
    TIMEOUT = 60

    def __init__(self, onCompletion):
        self.output = []
        self.error = []
        self.onCompletion = onCompletion
        self.onCompletion.addCallback(self.processOutput)

    def processOutput(self, output):
        return output

    def timeoutConnection(self):
        self.transport.signalProcess('KILL')

    def connectionMade(self):
        self.setTimeout(self.TIMEOUT)

    def outReceived(self, bytes):
        self.resetTimeout()
        self.output.append(bytes)

    def errReceived(self, bytes):
        self.resetTimeout()
        self.error.append(bytes)

    def processEnded(self, reason):
        self.setTimeout(None)
        if reason.check(error.ProcessTerminated):
            self.onCompletion.errback(ProcessFailed(self, reason))
        elif self.error:
            self.onCompletion.errback(ProcessFailed(self, None))
        else:
            self.onCompletion.callback(self.output)



class NoAttrsItem(item.Item):
    typeName = 'noattrsitem'
    schemaVersion = 1



class TransactedMethodItem(item.Item):
    """
    Helper class for testing the L{axiom.item.transacted} decorator.
    """
    value = text()
    calledWith = inmemory()

    def method(self, a, b, c):
        self.value = u"changed"
        self.calledWith = [a, b, c]
        raise Exception("TransactedMethodItem.method test exception")
    method.attribute = 'value'
    method = item.transacted(method)



class StoredNoticingItem(item.Item):
    """
    Test item which just remembers whether or not its C{stored} method has been
    called.
    """
    storedCount = integer(doc="""
    The number of times C{stored} has been called on this item.
    """, default=0)

    activatedCount = integer(doc="""
    The number of times C{stored} has been called on this item.
    """, default=0)

    activated = inmemory(doc="""
    A value set in the C{activate} callback and nowhere else.  Used to
    determine the ordering of C{activate} and C{stored} calls.
    """)

    def activate(self):
        self.activated = True


    def stored(self):
        self.storedCount += 1
        self.activatedCount += getattr(self, 'activated', 0)


class ItemWithDefault(item.Item):
    """
    Item with an attribute having a default value.
    """
    value = integer(default=10)



class ItemTestCase(unittest.TestCase):
    """
    Tests for L{Item}.
    """

    def test_repr(self):
        """
        L{Item.__repr__} should return a C{str} giving the name of the
        subclass and the names and values of all the item's attributes.
        """
        reprString = repr(ItemWithDefault(value=123))
        self.assertIn('value=123', reprString)
        self.assertIn('storeID=None', reprString)
        self.assertIn('ItemWithDefault', reprString)

        store = Store()
        item = ItemWithDefault(store=store, value=321)
        reprString = repr(item)
        self.assertIn('value=321', reprString)
        self.assertIn('storeID=%d' % (item.storeID,), reprString)
        self.assertIn('ItemWithDefault', reprString)


    def test_partiallyInitializedRepr(self):
        """
        L{Item.__repr__} should return a C{str} giving some information,
        even if called before L{Item.__init__} has run completely.
        """
        item = ItemWithDefault.__new__(ItemWithDefault)
        reprString = repr(item)
        self.assertIn('ItemWithDefault', reprString)


    def test_itemClassOrdering(self):
        """
        Test that L{Item} subclasses (not instances) sort by the Item's
        typeName.
        """
        A = TransactedMethodItem
        B = NoAttrsItem

        self.failUnless(A < B)
        self.failUnless(B >= A)
        self.failIf(A >= B)
        self.failIf(B <= A)
        self.failUnless(A != B)
        self.failUnless(B != A)
        self.failIf(A == B)
        self.failIf(B == A)


    def test_legacyItemComparison(self):
        """
        Legacy items with different versions must not compare equal.
        """
        legacy1 = declareLegacyItem('test_type', 1, {})
        legacy2 = declareLegacyItem('test_type', 2, {})
        self.assertNotEqual(legacy1, legacy2)
        self.assertEqual(legacy1, legacy1)
        self.assertEqual(legacy2, legacy2)


    def testCreateItem(self):
        st = store.Store()
        self.assertRaises(item.CantInstantiateItem, item.Item, store=st)


    def testCreateItemWithDefault(self):
        """
        Test that attributes with default values can be set to None properly.
        """
        st = store.Store()
        it = ItemWithDefault()
        it.value = None
        self.assertEqual(it.value, None)


    def test_storedCallbackAfterActivateCallback(self):
        """
        Test that L{Item.stored} is only called after L{Item.activate} has been
        called.
        """
        st = store.Store()
        i = StoredNoticingItem(store=st)
        self.assertEquals(i.activatedCount, 1)


    def test_storedCallbackOnAttributeSet(self):
        """
        Test that L{Item.stored} is called when an item is actually added to a
        store and not before.
        """
        st = store.Store()
        i = StoredNoticingItem()
        self.assertEquals(i.storedCount, 0)
        i.store = st
        self.assertEquals(i.storedCount, 1)


    def test_storedCallbackOnItemCreation(self):
        """
        Test that L{Item.stored} is called when an item is created with a
        store.
        """
        st = store.Store()
        i = StoredNoticingItem(store=st)
        self.assertEquals(i.storedCount, 1)


    def test_storedCallbackNotOnLoad(self):
        """
        Test that pulling an item out of a store does not invoke its stored
        callback again.
        """
        st = store.Store()
        storeID = StoredNoticingItem(store=st).storeID
        self.assertEquals(st.getItemByID(storeID).storedCount, 1)


    def testTransactedTransacts(self):
        """
        Test that a method wrapped in C{axiom.item.transacted} is automatically
        run in a transaction.
        """
        s = store.Store()
        i = TransactedMethodItem(store=s, value=u"unchanged")
        exc = self.assertRaises(Exception, i.method, 'a', 'b', 'c')
        self.assertEquals(exc.args, ("TransactedMethodItem.method test exception",))
        self.assertEquals(i.value, u"unchanged")


    def testTransactedPassedArguments(self):
        """
        Test that position and keyword arguments are passed through
        L{axiom.item.transacted}-wrapped methods correctly.
        """
        s = store.Store()
        i = TransactedMethodItem(store=s)
        exc = self.assertRaises(Exception, i.method, 'a', b='b', c='c')
        self.assertEquals(exc.args, ("TransactedMethodItem.method test exception",))
        self.assertEquals(i.calledWith, ['a', 'b', 'c'])


    def testTransactedPreservesAttributes(self):
        """
        Test that the original function attributes are available on a
        L{axiom.item.transacted}-wrapped function.
        """
        self.assertEquals(TransactedMethodItem.method.attribute, 'value')


    def testPersistentValues(self):
        st = store.Store()
        pi = itemtest.PlainItem(store=st, plain=u'hello')
        self.assertEqual(pi.persistentValues(), {'plain': u'hello'})


    def testPersistentValuesWithoutValue(self):
        st = store.Store()
        pi = itemtest.PlainItem(store=st)
        self.assertEqual(pi.persistentValues(), {'plain': None})


    def test_createItemWithNoAttrs(self):
        """
        Creating an item with no attributes succeeds.
        """
        st = store.Store()
        NoAttrsItem(store=st)


    def testCreatePlainItem(self):
        st = store.Store()
        s = itemtest.PlainItem(store=st)

    def testLoadLoadedPlainItem(self):
        """
        Test that loading an Item out of the store by its Store ID
        when a Python object representing that Item already exists in
        memory returns the same object as the one which already
        exists.
        """
        st = store.Store()
        item = itemtest.PlainItem(store=st)
        self.assertIdentical(item, st.getItemByID(item.storeID))

    def testLoadUnimportedPlainItem(self):
        """
        Test that an Item in the database can be loaded out of the
        database, even if the module defining its Python class has not
        been imported, as long as its class definition has not moved
        since it was added to the database.
        """
        storePath = filepath.FilePath(self.mktemp())
        st = store.Store(storePath)
        itemID = itemtest.PlainItem(store=st, plain=u'Hello, world!!!').storeID
        st.close()

        e = os.environ.copy()
        d = defer.Deferred()
        p = ProcessOutputCollector(d)
        try:
            reactor.spawnProcess(p, sys.executable, [sys.executable, '-Wignore', itemtestmain.__file__.rstrip('co'), storePath.path, str(itemID)], e)
        except NotImplementedError:
            raise unittest.SkipTest("Implement processes here")

        def cbOutput(output):
            self.assertEquals(''.join(output).strip(), 'Hello, world!!!')

        def ebBlah(err):
            log.err(err)
            self.fail(''.join(err.value.args[0].error))

        return d.addCallbacks(cbOutput, ebBlah)

    def testDeleteCreatePair(self):
        # Test coverage for a bug which was present in Axiom: deleting
        # the newest item in a database and then creating a new item
        # re-used the deleted item's oid causing all manner of
        # ridiculuousness.
        st = store.Store()

        i = itemtest.PlainItem(store=st)

        oldStoreID = i.storeID
        i.deleteFromStore()
        j = itemtest.PlainItem(store=st)
        self.failIfEqual(oldStoreID, j.storeID)

    def testDeleteThenLoad(self):
        st = store.Store()
        i = itemtest.PlainItem(store=st)
        oldStoreID = i.storeID
        self.assertEquals(st.getItemByID(oldStoreID, default=1234),
                          i)
        i.deleteFromStore()
        self.assertEquals(st.getItemByID(oldStoreID+100, default=1234),
                          1234)
        self.assertEquals(st.getItemByID(oldStoreID, default=1234),
                          1234)


    def test_duplicateDefinition(self):
        """
        When the same typeName is defined as an item class multiple times in
        memory, the second definition fails with a L{RuntimeError}.
        """
        class X(Item):
            dummy = integer()
        try:
            class X(Item):
                dummy = integer()
        except RuntimeError:
            pass
        else:
            self.fail("Duplicate definition should have failed.")


    def test_nonConflictingRedefinition(self):
        """
        If the python item class associated with a typeName is garbage
        collected, a new python item class can re-use that type name.
        """
        class X(Item):
            dummy = integer()
        del X
        class X(Item):
            dummy = integer()



class TestItem(Item):
    """
    Boring, behaviorless Item subclass used when we just need an item
    someplace.
    """
    attribute = integer()



class BrokenCommittedItem(Item):
    """
    Item class which changes database state in its committed method.  Don't
    write items like this, they're broken.
    """
    attribute = integer()
    _committed = inmemory()

    def committed(self):
        Item.committed(self)
        if getattr(self, '_committed', None) is not None:
            self._committed(self)



class CheckpointTestCase(TestCase):
    """
    Tests for Item checkpointing.
    """
    def setUp(self):
        self.checkpointed = []
        def checkpoint(item):
            self.checkpointed.append(item)
        self.originalCheckpoint = TestItem.checkpoint.im_func
        TestItem.checkpoint = checkpoint


    def tearDown(self):
        TestItem.checkpoint = self.originalCheckpoint


    def _autocommitBrokenCommittedMethodTest(self, method):
        store = Store()
        item = BrokenCommittedItem(store=store)
        item._committed = method
        self.assertRaises(ChangeRejected, setattr, item, 'attribute', 0)


    def _transactionBrokenCommittedMethodTest(self, method):
        store = Store()
        item = BrokenCommittedItem(store=store)
        item._committed = method

        def txn():
            item.attribute = 0
        self.assertRaises(ChangeRejected, store.transact, txn)


    def test_autocommitBrokenCommittedMethodMutate(self):
        """
        Test changing a persistent attribute in the committed (even if the
        original change was made in autocommit mode) callback raises
        L{ChangeRejected}.
        """
        def mutate(self):
            self.attribute = 0
        return self._autocommitBrokenCommittedMethodTest(mutate)


    def test_transactionBrokenCommittedMethodMutate(self):
        """
        Test changing a persistent attribute in the committed callback raises
        L{ChangeRejected}.
        """
        def mutate(item):
            item.attribute = 0
        return self._transactionBrokenCommittedMethodTest(mutate)


    def test_autocommitBrokenCommittedMethodDelete(self):
        """
        Test deleting an item in the committed (even if the original change was
        made in autocommit mode) callback raises L{ChangeRejected}.
        """
        def delete(item):
            item.deleteFromStore()
        return self._autocommitBrokenCommittedMethodTest(delete)


    def test_transactionBrokenCommittedMethodDelete(self):
        """
        Test changing a persistent attribute in the committed callback raises
        L{ChangeRejected}.
        """
        def delete(item):
            item.deleteFromStore()
        return self._transactionBrokenCommittedMethodTest(delete)


    def test_autocommitBrokenCommittedMethodCreate(self):
        """
        Test that creating a new item in a committed (even if the original
        change was made in autocommit mode) callback raises L{ChangeRejected}
        """
        def create(item):
            TestItem(store=item.store)
        return self._autocommitBrokenCommittedMethodTest(create)


    def test_transactionBrokenCommittedMethodCreate(self):
        """
        Test that creating a new item in a committed callback raises
        L{ChangeRejected}.
        """
        def create(item):
            TestItem(store=item.store)
        return self._transactionBrokenCommittedMethodTest(create)


    def test_autocommitCheckpoint(self):
        """
        Test that an Item is checkpointed when it is created outside of a
        transaction.
        """
        store = Store()
        item = TestItem(store=store)
        self.assertEquals(self.checkpointed, [item])


    def test_transactionCheckpoint(self):
        """
        Test that an Item is checkpointed when the transaction it is created
        within is committed.
        """
        store = Store()
        def txn():
            item = TestItem(store=store)
            self.assertEquals(self.checkpointed, [])
            return item
        item = store.transact(txn)
        self.assertEquals(self.checkpointed, [item])


    def test_queryCheckpoint(self):
        """
        Test that a newly created Item is checkpointed before a query is
        executed.
        """
        store = Store()
        def txn():
            item = TestItem(store=store)
            list(store.query(TestItem))
            self.assertEquals(self.checkpointed, [item])
        store.transact(txn)


    def test_autocommitTouchCheckpoint(self):
        """
        Test that an existing Item is checkpointed if it has an attribute
        changed on it.
        """
        store = Store()
        item = TestItem(store=store)

        # Get rid of the entry that's there from creation
        self.checkpointed = []

        item.attribute = 0
        self.assertEquals(self.checkpointed, [item])


    def test_transactionTouchCheckpoint(self):
        """
        Test that in a transaction an existing Item is checkpointed if it has
        touch called on it and the store it is in is checkpointed.
        """
        store = Store()
        item = TestItem(store=store)

        # Get rid of the entry that's there from creation
        self.checkpointed = []

        def txn():
            item.touch()
            store.checkpoint()
            self.assertEquals(self.checkpointed, [item])
        store.transact(txn)


    def test_twoQueriesOneCheckpoint(self):
        """
        Test that if two queries are performed in a transaction, a touched item
        only has checkpoint called on it before the first.
        """
        store = Store()
        item = TestItem(store=store)

        # Get rid of the entry that's there from creation
        self.checkpointed = []

        def txn():
            item.touch()
            list(store.query(TestItem))
            self.assertEquals(self.checkpointed, [item])
            list(store.query(TestItem))
            self.assertEquals(self.checkpointed, [item])
        store.transact(txn)



class TestInterface(Interface):
    """
    Testing interface.
    """



class TestInterface2(Interface):
    """
    Testing interface.
    """



class EmpowermentTests(SynchronousTestCase):
    """
    Tests for the C{@empowerment} class decorator.
    """
    def test_oneDecorator(self):
        """
        Decorating an item with C{@empowerment} declares it as empowering and
        implementing that interface.
        """
        @empowerment(TestInterface)
        class TI(Item):
            pass

        self.assertEqual(TI()._getPowerupInterfaces(), [(TestInterface, 0)])
        self.assertTrue(TestInterface.implementedBy(TI))


    def test_twoDecorators(self):
        """
        Decorating an item twice adds the powerup interfaces in the order of
        invocation.
        """
        @empowerment(TestInterface2, 20)
        @empowerment(TestInterface)
        class TI2(Item):
            pass

        self.assertEqual(
            TI2()._getPowerupInterfaces(),
            [(TestInterface, 0),
             (TestInterface2, 20)])
        self.assertTrue(TestInterface.implementedBy(TI2))
        self.assertTrue(TestInterface2.implementedBy(TI2))


    def test_decoratorAndAttribute(self):
        """
        Decorating an item with an existing C{powerupInterfaces} attribute adds
        the powerup interface to the end.
        """
        @empowerment(TestInterface)
        @implementer(TestInterface2)
        class TI3(Item):
            powerupInterfaces = [(TestInterface2, 20)]

        self.assertEqual(
            TI3()._getPowerupInterfaces(),
            [(TestInterface2, 20),
             (TestInterface, 0)])
        self.assertTrue(TestInterface.implementedBy(TI3))




import random

from axiom.item import Item
from axiom.attributes import text, timestamp, reference, integer, AND, OR
from axiom.store import Store
from epsilon import extime

_d = extime.Time.fromISO8601TimeAndDate

_books = [
    (u'Heart of Darkness', u'Joseph Conrad', u'0486264645', 80, _d('1990-07-01T00:00:00.000001')),
    (u'The Dark Tower, Book 7', u'Stephen King', u'1880418622', 864, _d('2004-11-21T00:00:00.000001')),
    (u'Guns, Germs, and Steel: The Fates of Human Societies', u'Jared Diamond', u'0393317552', 480, _d('1999-04-01T00:00:00.000001')),
    (u'The Lions of al-Rassan', u'Guy Gavriel Kay', u'0060733497', 528, _d('2005-06-28T00:00:00.000001')),
    ]

_borrowers = [u'Anne', u'Bob', u'Carol', u'Dave']


class Borrower(Item):
    typeName = 'borrower'
    schemaVersion = 1
    name = text(indexed=True)

class Book(Item):
    typeName = 'book'
    schemaVersion = 1

    title = text()
    author = text()
    isbn = text()
    pages = integer()
    datePublished = timestamp()

    lentTo = reference()
    library = reference()

class LendingLibrary(Item):
    typeName = 'lending_library'
    schemaVersion = 1

    name = text()

    def books(self):
        return self.store.query(Book,
                                Book.library == self)

    def getBorrower(self, name):
        for b in self.store.query(Borrower,
                                  Borrower.name == name):
            return b
        b = Borrower(name=name,
                     store=self.store)
        return b

    def initialize(self):
        for title, author, isbn, pages, published in _books:
            b = Book(
                title=title,
                author=author,
                isbn=isbn,
                pages=pages,
                datePublished=published,
                library=self,
                store=self.store)


    def displayBooks(self):
        for book in self.books():
            print book.title,
            if book.lentTo is not None:
                print 'lent to', '['+book.lentTo.name+']'
            else:
                print 'in library'

    def shuffleLending(self):
        for book in self.books():
            if book.lentTo is not None:
                print book.lentTo.name, 'returned', book.title
                book.lentTo = None
        for book in self.books():
            if random.choice([True, False]):
                borrower = random.choice(_borrowers)
                print 'Lending', book.title, 'to', borrower
                book.lentTo = self.getBorrower(borrower)

def main(s):
    for ll in s.query(LendingLibrary):
        print 'found existing library'
        break
    else:
        print 'creating new library'
        ll = LendingLibrary(store=s)
        ll.initialize()
    ll.displayBooks()
    print '***'
    ll.shuffleLending()
    print '---'
    ll.displayBooks()
    print '***'
    ll.shuffleLending()
    print '---'

    print s.count(Book, AND (Book.author == u'Stephen King',
                             Book.title == u'The Lions of al-Rassan'))
    print s.count(Book, OR (Book.author == u'Stephen King',
                            Book.title == u'The Lions of al-Rassan'))


if __name__ == '__main__':
    s = Store('testdb')
    s.transact(main, s)
    s.close()



from twisted.application.service import Service, IService
from twisted.python import filepath

from twisted.trial import unittest

from axiom.store import Store
from axiom.item import Item
from axiom.substore import SubStore

from axiom.attributes import text, bytes, boolean, inmemory



class SubStored(Item):
    schemaVersion = 1
    typeName = 'substoredthing'
    a = text()
    b = bytes()



class YouCantStartThis(Item, Service):
    parent = inmemory()
    running = inmemory()
    name = inmemory()

    started = boolean(default=False)

    def startService(self):
        self.started = True



class YouShouldStartThis(Item, Service):
    parent = inmemory()
    running = inmemory()
    name = inmemory()

    started = boolean(default=False)

    def startService(self):
        self.started = True



class SubStoreTest(unittest.TestCase):
    """
    Test on-disk creation of substores.
    """
    def testOneThing(self):
        """
        Ensure that items can be inserted into substores and
        subsequently retrieved.
        """
        topdb = filepath.FilePath(self.mktemp())
        s = Store(topdb)
        ss = SubStore.createNew(s, ['account', 'bob@divmod.com'])
        s2 = ss.open()

        ssd = SubStored(store=s2, a=u'hello world', b='what, its text')
        oid = ss.storeID
        oid2 = ssd.storeID

        s2.close()
        s.close()

        reopens = Store(topdb)
        reopenss = reopens.getItemByID(oid)
        reopens2 = reopenss.open()
        reopenssd = reopens2.getItemByID(oid2)

        self.assertEquals(reopenssd.a, u'hello world')
        self.assertEquals(reopenssd.b, 'what, its text')


    def test_oneThingMemory(self):
        """
        Ensure that items put into in-memory substores are retrievable.
        """
        s = Store()
        ss = SubStore.createNew(s, ['account', 'bob@divmod.com'])
        s2 = ss.open()

        ssd = SubStored(store=s2, a=u'hello world', b='what, its text')
        oid = ss.storeID
        oid2 = ssd.storeID

        s2.close()
        self.assertIdentical(s.getItemByID(oid), ss)
        self.assertIdentical(ss.open(), s2)
        item = s2.getItemByID(oid2)
        self.assertEquals(item.a, u'hello world')
        self.assertEquals(item.b, 'what, its text')


    def test_hereTodayGoneTomorrow(self):
        """
        Ensure that substores exist after closing them.
        """
        s = Store()
        ss = SubStore.createNew(s, ['account', 'bob@divmod.com'])
        s2 = ss.open()

        ssd = SubStored(store=s2, a=u'hello world', b='what, its text')
        oid = ss.storeID
        oid2 = ssd.storeID
        s2.close()
        #the following is done to mimic garbage collection of objects holding
        #on to substores
        del s2._openSubStore
        ss = s.getItemByID(oid)
        s2 = ss.open()
        item = s2.getItemByID(oid2)
        self.assertEquals(item.a, u'hello world')
        self.assertEquals(item.b, 'what, its text')


    def test_memorySubstoreFile(self):
        """
        In-memory substores whose stores have file directories should be able
        to create files.
        """
        filesdir = filepath.FilePath(self.mktemp())
        s = Store(filesdir=filesdir)
        ss = SubStore.createNew(s, ['account', 'bob@divmod.com'])
        s2 = ss.open()
        f = s2.newFile("test.txt")
        f.write("yay")
        f.close()
        self.assertEqual(open(f.finalpath.path).read(), "yay")


    def test_createNewStringPath(self):
        """
        Passing a string instead of a sequence of strings to
        L{SubStore.createNew} results in an exception.
        """
        s = Store()
        e = self.assertRaises(
            ValueError, SubStore.createNew, s, 'notasequence')
        self.assertEqual(
            e.args[0], "Received 'notasequence' instead of a sequence")


    def test_inheritParentConfiguration(self):
        """
        Substores use the debug and journal configuration of the parent store.
        """
        filesdir = filepath.FilePath(self.mktemp())
        s = Store(filesdir=filesdir, debug=True, journalMode=u'MEMORY')
        ss = SubStore.createNew(s, ['account', 'bob@divmod.com'])
        s2 = ss.open()
        self.assertEqual(s2.debug, True)
        self.assertEqual(s2.journalMode, u'MEMORY')



class SubStoreStartupSemantics(unittest.TestCase):
    """
    These tests verify that interactions between store and substore services
    are correct.  They also provide some documentation of expected edge-case
    behavior.  Read the code if you are interested in how to get startup
    notifications from substore items.
    """
    def setUp(self):
        """
        Set up the tests by creating a store and a substore and opening them
        both.
        """
        self.topdb = topdb = Store(filepath.FilePath(self.mktemp()))
        self.ssitem = ssitem = SubStore.createNew(
            topdb, ["dontstartme", "really"])
        self.ss = ssitem.open()
        self.serviceStarted = False

    def testDontStartNormally(self):
        """
        Substores' services are not supposed to be started when their parent
        stores are.
        """
        ss = self.ss
        ycst = YouCantStartThis(store=ss)
        ss.powerUp(ycst, IService)
        self._startService()
        self.failIf(ycst.started)

    def testStartEverythingExplicitly(self):
        """
        Substores implement IService themselves, just as regular stores do, via
        the special-case machinery.
        """
        ss = self.ss
        ysst = YouShouldStartThis(store=ss)
        ss.powerUp(ysst, IService)
        self.topdb.powerUp(self.ssitem, IService)
        self._startService()
        self.failUnless(ysst.started)

    def _startService(self):
        """
        Start the service and make sure we know it's started so tearDown can
        shut it down.
        """
        assert not self.serviceStarted
        self.serviceStarted = True
        return IService(self.topdb).startService()

    def tearDown(self):
        """
        Stop services that may have been started by these test cases.
        """
        if self.serviceStarted:
            return IService(self.topdb).stopService()




from twisted.cred.portal import Portal, IRealm

from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.credentials import UsernamePassword

from axiom.test.test_userbase import IGarbage

from axiom.test.historic import stubloader

SECRET = 'asdf'
SECRET2 = 'ghjk'

class AccountUpgradeTest(stubloader.StubbedTest):
    def testUpgrade(self):
        p = Portal(IRealm(self.store),
                   [ICredentialsChecker(self.store)])

        def loggedIn((ifc, av, lgo)):
            assert av.garbage == 7
            # Bug in cooperator?  this triggers an exception.
            # return svc.stopService()
        d = p.login(
            UsernamePassword('test@example.com', SECRET), None, IGarbage)
        return d.addCallback(loggedIn)



# -*- test-case-name: axiom.test.test_upgrading -*-

"""
Axiom Item/schema upgrade support.
"""
from twisted.python.failure import Failure
from twisted.python.log import msg
from twisted.python.reflect import qual

from axiom.errors import NoUpgradePathAvailable, UpgraderRecursion
from axiom.errors import ItemUpgradeError
from axiom.item import _legacyTypes, _typeNameToMostRecentClass
from axiom._schema import (
    CREATE_OBJECTS, CREATE_OBJECTS_IDX, CREATE_TYPES, LATEST_TYPES)


_upgradeRegistry = {}


class _StoreUpgrade(object):
    """
    Manage Item upgrades and upgrade batching for a store.

    @type _currentlyUpgrading: C{dict}
    @ivar _currentlyUpgrading: A map of storeIDs to Items currently in the
        middle of an upgrader.  Used to make sure that the same item isn't
        upgraded reentrantly.

    @type _oldTypesRemaining: C{list}
    @ivar _oldTypesRemaining: All the old types which have not been fully
        upgraded in this database.
    """

    def __init__(self, store):
        self.store = store
        self._currentlyUpgrading = {}
        self._oldTypesRemaining = []


    def upgradesPending(self):
        return bool(self._oldTypesRemaining)

    upgradesPending = property(
        upgradesPending,
        doc="""
        Flag indicating whether there any types that still need to be upgraded
        or not.
        """)


    def checkUpgradePaths(self):
        """
        Check that all of the accumulated old Item types have a way to get
        from their current version to the latest version.

        @raise axiom.errors.NoUpgradePathAvailable: for any, and all, Items
            that do not have a valid upgrade path
        """
        cantUpgradeErrors = []

        for oldVersion in self._oldTypesRemaining:
            # We have to be able to get from oldVersion.schemaVersion to
            # the most recent type.

            currentType = _typeNameToMostRecentClass.get(
                oldVersion.typeName, None)

            if currentType is None:
                # There isn't a current version of this type; it's entirely
                # legacy, will be upgraded by deleting and replacing with
                # something else.
                continue

            typeInQuestion = oldVersion.typeName
            upgver = oldVersion.schemaVersion

            while upgver < currentType.schemaVersion:
                # Do we have enough of the schema present to upgrade?
                if ((typeInQuestion, upgver)
                    not in _upgradeRegistry):
                    cantUpgradeErrors.append(
                        "No upgrader present for %s (%s) from %d to %d" % (
                            typeInQuestion, qual(currentType), upgver,
                            upgver + 1))

                # Is there a type available for each upgrader version?
                if upgver+1 != currentType.schemaVersion:
                    if (typeInQuestion, upgver+1) not in _legacyTypes:
                        cantUpgradeErrors.append(
                            "Type schema required for upgrade missing:"
                            " %s version %d" % (
                                typeInQuestion, upgver+1))
                upgver += 1

            if cantUpgradeErrors:
                raise NoUpgradePathAvailable('\n    '.join(cantUpgradeErrors))


    def queueTypeUpgrade(self, oldtype):
        """
        Queue a type upgrade for C{oldtype}.
        """
        if oldtype not in self._oldTypesRemaining:
            self._oldTypesRemaining.append(oldtype)


    def upgradeItem(self, thisItem):
        """
        Upgrade a legacy item.

        @raise axiom.errors.UpgraderRecursion: If the given item is already in
            the process of being upgraded.
        """
        sid = thisItem.storeID
        if sid in self._currentlyUpgrading:
            raise UpgraderRecursion()
        self._currentlyUpgrading[sid] = thisItem
        try:
            return upgradeAllTheWay(thisItem)
        finally:
            self._currentlyUpgrading.pop(sid)


    def upgradeEverything(self):
        """
        Upgrade every item in the store, one at a time.

        @raise axiom.errors.ItemUpgradeError: if an item upgrade failed

        @return: A generator that yields for each item upgrade.
        """
        return self.upgradeBatch(1)


    def upgradeBatch(self, n):
        """
        Upgrade the entire store in batches, yielding after each batch.

        @param n: Number of upgrades to perform per transaction
        @type n: C{int}

        @raise axiom.errors.ItemUpgradeError: if an item upgrade failed

        @return: A generator that yields after each batch upgrade. This needs
            to be consumed for upgrading to actually take place.
        """
        store = self.store

        def _doBatch(itemType):
            upgradedAnything = False

            for theItem in store.query(itemType, limit=n):
                upgradedAnything = True
                try:
                    self.upgradeItem(theItem)
                except:
                    f = Failure()
                    raise ItemUpgradeError(
                        f, theItem.storeID, itemType,
                        _typeNameToMostRecentClass[itemType.typeName])

            return upgradedAnything

        if self.upgradesPending:
            didAny = False

            while self._oldTypesRemaining:
                t0 = self._oldTypesRemaining[0]

                upgradedAnything = store.transact(_doBatch, t0)
                if not upgradedAnything:
                    self._oldTypesRemaining.pop(0)
                    if didAny:
                        msg("%s finished upgrading %s" % (store.dbdir.path, qual(t0)))
                    continue
                elif not didAny:
                    didAny = True
                    msg("%s beginning upgrade..." % (store.dbdir.path,))

                yield None

            if didAny:
                msg("%s completely upgraded." % (store.dbdir.path,))



def registerUpgrader(upgrader, typeName, oldVersion, newVersion):
    """
    Register a callable which can perform a schema upgrade between two
    particular versions.

    @param upgrader: A one-argument callable which will upgrade an object.  It
    is invoked with an instance of the old version of the object.
    @param typeName: The database typename for which this is an upgrader.
    @param oldVersion: The version from which this will upgrade.
    @param newVersion: The version to which this will upgrade.  This must be
    exactly one greater than C{oldVersion}.
    """
    # assert (typeName, oldVersion, newVersion) not in _upgradeRegistry, "duplicate upgrader"
    # ^ this makes the tests blow up so it's just disabled for now; perhaps we
    # should have a specific test mode
    # assert newVersion - oldVersion == 1, "read the doc string"
    assert isinstance(typeName, str), "read the doc string"
    _upgradeRegistry[typeName, oldVersion] = upgrader

def registerAttributeCopyingUpgrader(itemType, fromVersion, toVersion, postCopy=None):
    """
    Register an upgrader for C{itemType}, from C{fromVersion} to C{toVersion},
    which will copy all attributes from the legacy item to the new item.  If
    postCopy is provided, it will be called with the new item after upgrading.

    @param itemType: L{axiom.item.Item} subclass
    @param postCopy: a callable of one argument
    @return: None
    """
    def upgrader(old):
        newitem = old.upgradeVersion(itemType.typeName, fromVersion, toVersion,
                                     **dict((str(name), getattr(old, name))
                                            for (name, _) in old.getSchema()))
        if postCopy is not None:
            postCopy(newitem)
        return newitem
    registerUpgrader(upgrader, itemType.typeName, fromVersion, toVersion)


def registerDeletionUpgrader(itemType, fromVersion, toVersion):
    """
    Register an upgrader for C{itemType}, from C{fromVersion} to C{toVersion},
    which will delete the item from the database.

    @param itemType: L{axiom.item.Item} subclass
    @return: None
    """
    # XXX This should actually do something more special so that a new table is
    # not created and such.
    def upgrader(old):
        old.deleteFromStore()
        return None
    registerUpgrader(upgrader, itemType.typeName, fromVersion, toVersion)


def upgradeAllTheWay(o):
    assert o.__legacy__
    while True:
        try:
            upgrader = _upgradeRegistry[o.typeName, o.schemaVersion]
        except KeyError:
            break
        else:
            o = upgrader(o)
            if o is None:
                # Object was explicitly destroyed during upgrading.
                break
    return o



def _hasExplicitOid(store, table):
    """
    Does the given table have an explicit oid column?
    """
    return any(info[1] == 'oid' for info
               in store.querySchemaSQL(
                   'PRAGMA *DATABASE*.table_info({})'.format(table)))



def _upgradeTableOid(store, table, createTable, postCreate=lambda: None):
    """
    Upgrade a table to have an explicit oid.

    Must be called in a transaction to avoid corrupting the database.
    """
    if _hasExplicitOid(store, table):
        return
    store.executeSchemaSQL(
        'ALTER TABLE *DATABASE*.{0} RENAME TO {0}_temp'.format(table))
    createTable()
    store.executeSchemaSQL(
        'INSERT INTO *DATABASE*.{0} '
        'SELECT oid, * FROM *DATABASE*.{0}_temp'.format(table))
    store.executeSchemaSQL('DROP TABLE *DATABASE*.{0}_temp'.format(table))
    postCreate()



def upgradeSystemOid(store):
    """
    Upgrade the system tables to use explicit oid columns.
    """
    store.transact(
        _upgradeTableOid, store, 'axiom_types',
        lambda: store.executeSchemaSQL(CREATE_TYPES))
    store.transact(
        _upgradeTableOid, store, 'axiom_objects',
        lambda: store.executeSchemaSQL(CREATE_OBJECTS),
        lambda: store.executeSchemaSQL(CREATE_OBJECTS_IDX))



def upgradeExplicitOid(store):
    """
    Upgrade a store to use explicit oid columns.
    
    This allows VACUUMing the database without corrupting it.

    This requires copying all of axiom_objects and axiom_types, as well as all
    item tables that have not yet been upgraded.  Consider VACUUMing the
    database afterwards to reclaim space.
    """
    upgradeSystemOid(store)
    for typename, version in store.querySchemaSQL(LATEST_TYPES):
        cls = _typeNameToMostRecentClass[typename]
        if cls.schemaVersion != version:
            remaining = store.querySQL(
                'SELECT oid FROM {} LIMIT 1'.format(
                    store._tableNameFor(typename, version)))
            if len(remaining) == 0:
                # Nothing to upgrade
                continue
            else:
                raise RuntimeError(
                    '{}:{} not fully upgraded to {}'.format(
                        typename, version, cls.schemaVersion))
        store.transact(
            _upgradeTableOid,
            store,
            store._tableNameOnlyFor(typename, version),
            lambda: store._justCreateTable(cls),
            lambda: store._createIndexesFor(cls, []))



__all__ = [
    'registerUpgrader', 'registerAttributeCopyingUpgrader',
    'registerDeletionUpgrader', 'upgradeExplicitOid']



# Copyright 2008 Divmod, Inc.
# See LICENSE file for details

"""
Tests for Axiom store version history.
"""

import sys, StringIO
from twisted.trial import unittest
from twisted.python.versions import Version

from axiom.store import Store
from axiom import version as axiom_version
from axiom.listversions import (getSystemVersions,
                                SystemVersion,
                                checkSystemVersion)

from axiom.scripts.axiomatic import Options as AxiomaticOptions
from axiom.test.util import CommandStubMixin
from axiom.plugins.axiom_plugins import ListVersions

class SystemVersionTests(unittest.TestCase, CommandStubMixin):
    """
    Tests for recording the versions of software used to open a store
    throughout its lifetime.
    """

    def setUp(self):
        """
        Create an on-disk store.
        """
        self.dbdir = self.mktemp()
        self.store = Store(self.dbdir)


    def _reopenStore(self):
        """
        Close the store and reopen it.
        """
        self.store.close()
        self.store = Store(self.dbdir)


    def test_getSystemVersions(self):
        """
        L{getSystemVersions} returns all the version plugins it finds.
        """
        someVersions = [Version("foo", 1, 2, 3),
                        Version("baz", 0, 0, 1)]
        def getSomeVersions(iface, package):
            return someVersions
        self.assertEqual(getSystemVersions(getSomeVersions),
                         someVersions)

    def test_checkSystemVersion(self):
        """
         Calling checkSystemVersion:
            1. Doesn't duplicate the system version when called with the
               same software package versions.
            2. Creates a new system version when one of the software
               package versions has changed.
            3. Notices and creates a new system version when the system
               config has reverted to a previous state.
        """
        versions = [Version("foo", 1, 2, 3)]

        checkSystemVersion(self.store, versions)
        checkSystemVersion(self.store, versions)

        query_results = list(self.store.query(SystemVersion))
        self.assertEquals(len(query_results), 1)

        # Adjust a version number and try again.
        v = versions[0]
        versions[0] = Version(v.package, v.major, v.minor + 1, v.micro)
        checkSystemVersion(self.store, versions)

        query_results = list(self.store.query(SystemVersion))
        self.assertEquals(len(query_results), 2)

        # Revert the version number and try again.
        versions[0] = v

        checkSystemVersion(self.store, versions)
        query_results = list(self.store.query(SystemVersion))
        self.assertEquals(len(query_results), 3)

        # Reopening the store does not duplicate the version.
        self._reopenStore()
        query_results = list(self.store.query(SystemVersion))
        self.assertEquals(len(query_results), 3)


    def test_commandLine(self):
        """
        L{ListVersions} will list versions of code used in this store when
        invoked as an axiomatic subcommand.
        """
        checkSystemVersion(self.store)

        out = StringIO.StringIO()
        self.patch(sys, 'stdout', out)
        lv = ListVersions()
        lv.parent = self
        lv.parseOptions([])
        result = out.getvalue()
        self.assertSubstring("axiom: " + axiom_version.short(), result)


    def test_axiomaticSubcommand(self):
        """
        L{ListVersions} is available as a subcommand of I{axiomatic}.
        """
        subCommands = AxiomaticOptions().subCommands
        [options] = [cmd[2] for cmd in subCommands if cmd[0] == 'list-version']
        self.assertIdentical(options, ListVersions)



# Copyright 2009 Divmod, Inc.  See LICENSE file for details

"""
Helper for axiomatic reactor-selection unit tests.
"""

# The main point of this file: import the default reactor.
from twisted.internet import reactor

# Define an Item, too, so that it can go into a Store and trigger an import of
# this module at schema-check (ie, store opening) time.
from axiom.item import Item
from axiom.attributes import integer

class SomeItem(Item):
    attribute = integer()





from testbase import con, cur

cur.execute("create table foo (bar int, baz varchar)")

for x in range(500):
    cur.execute("insert into foo values (?, ?)",
                (x, "string-value-of-"+str(x)))

con.commit()



import gc

from twisted.trial.unittest import TestCase

from axiom.store import Store
from axiom.upgrade import registerUpgrader, registerAttributeCopyingUpgrader
from axiom.item import Item, declareLegacyItem
from axiom.attributes import integer, reference
from axiom.errors import BrokenReference, DeletionDisallowed

class Referee(Item):
    schemaVersion = 1
    typeName = "test_reference_referee"

    topSecret = integer()

class SimpleReferent(Item):
    schemaVersion = 1
    typeName = "test_reference_referent"

    ref = reference()


class BreakingReferent(Item):
    schemaVersion = 1
    typeName = "test_reference_breaking_referent"

    ref = reference(whenDeleted=reference.NULLIFY)

class DependentReferent(Item):
    ref = reference(whenDeleted=reference.CASCADE, reftype=Referee)

class DisallowReferent(Item):
    ref = reference(whenDeleted=reference.DISALLOW, reftype=Referee)

class BadReferenceTestCase(TestCase):
    ntimes = 10

    def testSanity(self):
        store = Store()
        for i in xrange(self.ntimes):
            SimpleReferent(store=store, ref=Referee(store=store, topSecret=i))
            (referee,) = list(store.query(Referee))
            (referent,) = list(store.query(SimpleReferent))
            self.assertEqual(referent.ref.topSecret, referee.topSecret)
            referee.deleteFromStore()
            referent.deleteFromStore()

    def testBadReferenceNone(self):
        """
        Test that accessing a broken reference on an Item that has already been
        loaded into memory correctly nullifies the attribute.
        """
        store = Store()
        referee = Referee(store=store, topSecret=0)
        referent = SimpleReferent(store=store, ref=referee)
        referee.deleteFromStore()

        referee = None
        gc.collect()

        (referent,) = list(store.query(SimpleReferent))
        self.assertEqual(referent.ref, None)

    def testBadReferenceNoneLoading(self):
        """
        Test that accessing a broken reference on an Item that has not yet been
        loaded correctly nullifies the attribute.
        """
        store = Store()
        referee = Referee(store=store, topSecret=0)
        referent = SimpleReferent(store=store, ref=referee)
        referee.deleteFromStore()

        referee = None
        referent = None
        gc.collect()

        (referent,) = list(store.query(SimpleReferent))
        self.assertEqual(referent.ref, None)


    def test_brokenReferenceException(self):
        """
        Test that an exception is raised when a broken reference is detected
        when this should be impossible (ie. CASCADE or NULLIFY).
        """
        store = Store()

        referee = Referee(store=store, topSecret=0)
        referent = BreakingReferent(store=store, ref=referee)

        referee.deleteFromStore()
        referent = None
        gc.collect()

        referent = store.findFirst(BreakingReferent)
        self.patch(BreakingReferent.ref, 'whenDeleted', reference.CASCADE)
        self.assertRaises(BrokenReference, lambda: referent.ref)


    def testBadReferenceNoneRevert(self):
        store = Store()
        referee = Referee(store=store, topSecret=0)
        referent = SimpleReferent(store=store, ref=referee)
        def txn():
            referee.deleteFromStore()
            self.assertEqual(referent.ref, None)
            1 / 0
        self.assertRaises(ZeroDivisionError, store.transact, txn)
        self.assertEqual(referent.ref, referee)

        referent = None
        referee = None
        gc.collect()

        referent = store.findUnique(SimpleReferent)
        referee = store.findUnique(Referee)
        self.assertEqual(referent.ref, referee)

    def testBrokenReferenceDisallow(self):
        """
        Test that deleting an item referred to by a whenDeleted == DISALLOW
        reference raises an exception.
        """
        store = Store()
        referee = Referee(store=store, topSecret=0)
        referent = DisallowReferent(store=store, ref=referee)

        self.assertRaises(DeletionDisallowed, referee.deleteFromStore)
        self.assertRaises(DeletionDisallowed, store.query(Referee).deleteFromStore)

    def testReferenceQuery(self):
        store = Store()
        referee = Referee(store=store, topSecret=0)
        self.assertEqual(
            list(store.query(SimpleReferent,
                             SimpleReferent.ref == Referee.storeID)),
            [])

    def testReferenceDeletion(self):
        store = Store()
        referee = Referee(store=store, topSecret=0)
        dep = DependentReferent(store=store,
                                ref=referee)
        sid = dep.storeID
        self.assertIdentical(store.getItemByID(sid), dep) # sanity
        referee.deleteFromStore()
        self.assertRaises(KeyError, store.getItemByID, sid)

    def testBatchReferenceDeletion(self):
        """
        Test that batch deletion removes dependent items correctly.
        """
        store = Store()
        referee = Referee(store=store, topSecret=0)
        dep = DependentReferent(store=store,
                                ref=referee)
        sid = dep.storeID
        store.query(Referee).deleteFromStore()
        self.assertRaises(KeyError, store.getItemByID, sid)


    def test_dummyItemReference(self):
        """
        Getting the value of a reference attribute which has previously been
        set to a legacy item results in an instance of the most recent type for
        that item.
        """
        store = Store()
        referent = SimpleReferent(store=store)
        oldReferee = nonUpgradedItem(store=store)
        referent.ref = oldReferee
        newReferee = referent.ref
        self.assertTrue(
            isinstance(newReferee, UpgradedItem),
            "%r was instance of %r, expected %r" % (newReferee,
                                                    type(newReferee),
                                                    UpgradedItem))


    def test_dummyItemReferenceUpgraded(self):
        """
        Getting the value of a reference attribute which has been set to a
        legacy item, which is then upgraded while the reference is "live",
        results in an instance of the most recent type for that item.
        """
        store = Store()
        referent = SimpleReferent(store=store)
        oldReferee = nonUpgradedItem2(store=store)
        referent.ref = oldReferee
        # Manually run the upgrader on this specific legacy item. This is the
        # same as if the SimpleReferent item had been created in an upgrader
        # for UpgradedItem, except that we can keep a strong reference to
        # oldReferee to ensure it is not garbage collected (this would
        # otherwise happen nondeterministically on platforms like PyPy).
        newReferee = item2to3(oldReferee)
        self.assertIsInstance(newReferee, UpgradedItem)
        self.assertIdentical(referent.ref, newReferee)


    def test_dummyItemReferenceInUpgrade(self):
        """
        Setting the value of a reference attribute to a legacy item during an
        upgrade results in the same value being set on the upgraded item.
        """
        store = Store()
        def tx():
            oldReferent = nonUpgradedItem(store=store)
            oldReferee = nonUpgradedItem(store=store)
            newReferent = oldReferent.upgradeVersion(
                UpgradedItem.typeName, 1, 2)
            newReferee = oldReferee.upgradeVersion(
                UpgradedItem.typeName, 1, 2, ref=newReferent)
            self.assertIdentical(newReferee.ref, newReferent)
        store.transact(tx)


    def test_dummyItemGetItemByID(self):
        """
        Instantiating a dummy item and then getting it by its storeID should
        upgrade it.
        """
        store = Store()
        t = nonUpgradedItem(store=store)
        self.assertEquals(t.__legacy__, True)
        self.assertRaises(KeyError, store.objectCache.get, t.storeID)
        t2 = store.getItemByID(t.storeID)
        self.assertNotIdentical(t, t2)
        self.assertTrue(isinstance(t2, UpgradedItem))



class UpgradedItem(Item):
    """
    A simple item which is the current version of L{nonUpgradedItem}.
    """
    schemaVersion = 3
    ref = reference()



nonUpgradedItem = declareLegacyItem(
    UpgradedItem.typeName, 1,
    dict(ref=reference()))



nonUpgradedItem2 = declareLegacyItem(
    UpgradedItem.typeName, 2,
    dict(ref=reference()))

registerAttributeCopyingUpgrader(UpgradedItem, 1, 2)


def item2to3(old):
    """
    Upgrade an nonUpgradedItem to UpgradedItem
    """
    return old.upgradeVersion(UpgradedItem.typeName, 2, 3, ref=old.ref)

registerUpgrader(item2to3, UpgradedItem.typeName, 2, 3)



# -*- test-case-name: axiom.test.test_upgrading.DeletionTest.testCircular -*-
from axiom.item import Item

from axiom.attributes import reference, integer

class A(Item):
    typeName = 'test_circular_a'
    b = reference()

class B(Item):
    typeName = 'test_circular_b'
    a = reference()
    n = integer()




# -*- test-case-name: axiom.test.test_upgrading.DuringUpgradeTests.test_reentrantUpgraderFailure -*-

from axiom.attributes import integer, reference
from axiom.item import Item, normalize
from axiom.upgrade import registerUpgrader

class Simple(Item):
    # Don't import the old schema. -exarkun
    typeName = normalize(
        "axiom.test.upgrade_fixtures.reentrant_old.Simple")
    schemaVersion = 2
    dummy = integer()
    selfReference = reference()

def upgradeSimple1to2(old):
    # Force the upgrade.
    selfRef = old.store.getItemByID(old.storeID)
    return old.upgradeVersion(
        old.typeName, 1, 2,
        dummy=old.dummy,
        selfReference=selfRef)

registerUpgrader(upgradeSimple1to2, Simple.typeName, 1, 2)



# -*- test-case-name: axiom.test.test_upgrading.DuringUpgradeTests.test_referenceModifiedByForeignUpgrader -*-

from axiom.attributes import reference, integer
from axiom.item import Item, normalize
from axiom.upgrade import registerUpgrader

NEW_VALUE = 71

class Referrer(Item):
    """
    An item which just refers to another kind of item which will be upgraded.
    """
    # Don't import the old schema. -exarkun
    typeName = normalize(
        "axiom.test.upgrade_fixtures.replace_attribute_old.Referrer")
    referee = reference()


class Referee(Item):
    """
    An item the upgrader of which replaces itself on L{Referrer} with a new
    instance with a different value.
    """
    # Don't import the old schema. -exarkun
    typeName = normalize(
        "axiom.test.upgrade_fixtures.replace_attribute_old.Referee")
    schemaVersion = 2
    value = integer()


def referee1to2(oldReferee):
    """
    Find the L{Referrer} which refers to C{oldReferee} and replace its
    C{referee} attribute with a new, different L{Referee} item with a different
    C{value}.
    """
    store = oldReferee.store
    [referrer] = list(store.query(Referrer, Referrer.referee == oldReferee))
    referrer.referee = Referee(store=store, value=NEW_VALUE)
    return oldReferee.upgradeVersion(
        Referee.typeName, 1, 2, value=oldReferee.value)

registerUpgrader(referee1to2, Referee.typeName, 1, 2)
